from utils import (
    addition,
    skip_if_no_anthropic,
    skip_if_no_google,
    skip_if_no_mistral,
    skip_if_no_openai,
)

from inspect_ai import Task, eval
from inspect_ai.dataset import Sample
from inspect_ai.log import EvalLog
from inspect_ai.model import (
    ChatMessage,
    ChatMessageAssistant,
    ChatMessageTool,
    Model,
    ToolCall,
    ToolFunction,
    get_model,
)
from inspect_ai.scorer import match
from inspect_ai.solver import generate, tool, use_tools

# we define 3 versions of addition so we can test the ability to force the
# the model to use a certain tool via tool_choice=ToolFunction()


# define some other tools to test forcing tool usage
@tool(
    prompt="""
    If you are given a math problem of any kind,
    please use the addition tool to compute the result.
"""
)
def addition2():
    async def add(x: int, y: int):
        """
        Tool for adding two numbers.

        Args:
            x (int): First number to add.
            y (int): Second number to add.

        Returns:
            The sum of the two numbers.
        """
        return x + y

    return add


# define some other tools to test forcing tool usage
@tool(
    prompt="""
    If you are given a math problem of any kind,
    please use the addition tool to compute the result.
"""
)
def addition3():
    async def add(x: int, y: int):
        """
        Tool for adding two numbers.

        Args:
            x (int): First number to add.
            y (int): Second number to add.

        Returns:
            The sum of the two numbers.
        """
        return x + y

    return add


def check_tools(model: Model, **model_args) -> None:
    check_tools_calls(model, **model_args)
    check_tools_none(model, **model_args)
    check_tools_force(model, **model_args)


addition_dataset = [
    Sample(
        input="What is 1 + 1?", target=["2", "2.0", "Two"], metadata={"color": "red"}
    )
]


def check_tools_calls(model: Model, **model_args) -> None:
    model = get_model(model)
    task = Task(
        dataset=addition_dataset,
        plan=[use_tools(addition()), generate()],
        scorer=match(),
    )

    # evaluate the task
    log: list[EvalLog] = eval(task, model=model, model_args=model_args)

    # check that we got the answer right
    assert log[0].results and log[0].results.metrics["accuracy"].value == 1

    # check that there is a tool_call
    assert log[0].samples
    messages = log[0].samples[0].messages
    tool_call = get_tool_call(messages, "addition")
    assert tool_call

    # check that there is a tool response for this call
    assert get_tool_response(messages, tool_call)


def check_tools_none(model: Model, **model_args) -> None:
    model = get_model(model)
    task = Task(
        dataset=addition_dataset,
        plan=[use_tools(addition(), tool_choice="none"), generate()],
        scorer=match(),
    )

    # evaluate the task
    log: list[EvalLog] = eval(task, model=model, model_args=model_args)

    # confirm no tool calls
    assert log[0].samples
    messages = log[0].samples[0].messages
    tool_call = get_tool_call(messages, "addition")
    assert tool_call is None


def check_tools_force(model: Model, **model_args) -> None:
    model = get_model(model)
    task = Task(
        dataset=addition_dataset,
        plan=[
            use_tools(
                [addition(), addition2(), addition3()],
                tool_choice=ToolFunction(name="addition2"),
            ),
            generate(),
        ],
        scorer=match(),
    )

    # evaluate the task
    log: list[EvalLog] = eval(task, model=model, model_args=model_args)

    # confirm we called the right tool
    assert log[0].samples
    messages = log[0].samples[0].messages
    tool_call = get_tool_call(messages, "addition2")
    assert tool_call is not None and tool_call.function == "addition2"


@skip_if_no_openai
def test_openai_tools():
    check_tools("openai/gpt-4")


@skip_if_no_anthropic
def test_anthropic_tools():
    check_tools("anthropic/claude-3-sonnet-20240229", tools_beta=False)
    check_tools("anthropic/claude-3-sonnet-20240229", tools_beta=True)


@skip_if_no_mistral
def test_mistral_tools():
    check_tools("mistral/mistral-large-latest")


@skip_if_no_google
def test_google_tools():
    check_tools("google/gemini-1.0-pro")


def get_tool_call(messages: list[ChatMessage], tool: str) -> ToolCall | None:
    assistant_messages = [
        message for message in messages if isinstance(message, ChatMessageAssistant)
    ]
    tool_call_message = next(
        (
            message
            for message in assistant_messages
            if message.tool_calls and len(message.tool_calls)
        ),
        None,
    )
    if tool_call_message:
        return next(
            (
                tool_call
                for tool_call in (tool_call_message.tool_calls or [])
                if tool_call.function == tool
            ),
            None,
        )
    else:
        return None


def get_tool_response(messages: list[ChatMessage], tool_call: ToolCall) -> str | None:
    tool_messages = [
        message for message in messages if isinstance(message, ChatMessageTool)
    ]
    tool_response = next(
        (message for message in tool_messages if message.tool_call_id == tool_call.id),
        None,
    )
    if tool_response:
        return tool_response.text
    else:
        return None



import shutil
import subprocess

from pydantic import BaseModel

from .path import chdir


class GitContext(BaseModel):
    origin: str
    commit: str


def git_context(dir: str) -> GitContext | None:
    with chdir(dir):
        # check for git
        git = shutil.which("git")
        if not git:
            return None

        # check for a git revision in this directory
        commit_result = subprocess.run(
            [git, "rev-parse", "--short", "HEAD"], capture_output=True, text=True
        )
        if commit_result.returncode != 0:
            return None

        # check for git origin (if any)
        origin = subprocess.run(
            [git, "remote", "get-url", "origin"],
            capture_output=True,
            text=True,
        ).stdout.strip()

        # return context
        return GitContext(origin=origin, commit=commit_result.stdout.strip())



import pytest
from utils import skip_if_no_openai, skip_if_no_together

from inspect_ai.model import ChatMessageUser, GenerateConfig, ModelOutput, get_model


async def generate_with_logprobs(model_name) -> ModelOutput:
    model = get_model(
        model_name,
        config=GenerateConfig(logprobs=True, top_logprobs=2),
    )

    message = ChatMessageUser(content="Hello.")
    return await model.generate(input=[message])


@pytest.mark.asyncio
@skip_if_no_openai
async def test_openai_logprobs() -> None:
    response = await generate_with_logprobs("openai/gpt-3.5-turbo")
    assert response.choices[0].logprobs is not None
    assert len(response.choices[0].logprobs["content"][0]["top_logprobs"]) == 2


@pytest.mark.asyncio
@skip_if_no_together
async def test_together_logprobs() -> None:
    response = await generate_with_logprobs("together/lmsys/vicuna-13b-v1.5")
    assert (
        response.choices[0].logprobs
        and response.choices[0].logprobs["token_ids"] is not None
    )



import re
from typing import (
    Any,
    Callable,
    Protocol,
    Tuple,
    TypeVar,
    cast,
    runtime_checkable,
)

from inspect_ai._util.registry import (
    RegistryInfo,
    registry_add,
    registry_name,
    registry_tag,
)

ToolResult = str | int | float | bool | Tuple[str | int | float | bool, dict[str, Any]]


@runtime_checkable
class Tool(Protocol):
    async def __call__(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> ToolResult:
        r"""Additional tool that an agent can use to solve a task.

        Args:
            *args (Any): Arguments for the tool.
            **kwargs (Any): Keyword arguments for the tool.

        Returns:
            Single value or a tuple containing the value and
            metadata to add to the task state
        """
        ...


ToolType = TypeVar("ToolType", Callable[..., Tool], type[Tool])
r"""Tool type.

Valid tool types include:
 - Functions that return a Tool
 - Classes derivied from Tool
"""


def tool_register(tool: ToolType, name: str) -> ToolType:
    r"""Register a function or class as a tool.

    Args:
        tool (ToolType):
            Tool function or a class derived from Tool.
        docstring (Docstring): Docstring for the tool. Used to extract arg descriptions.
        name (str): Name of tool (Optional, defaults to object name)

    Returns:
        Tool with registry attributes.
    """
    registry_add(
        tool,
        RegistryInfo(type="tool", name=name),
    )
    return tool


def tool(
    prompt: str | None = None,
    params: dict[str, str] = {},
    name: str | None = None,
) -> Callable[[Callable[..., Tool]], Callable[..., Tool]]:
    r"""Decorator for registering tools.

    Args:
        prompt (str):
            System prompt associated with this tool (provides
            guideance to the LLM on how to use the tool)
        name (str | None):
            Optional name for tool. If the decorator has no name
            argument then the name of the underlying ToolType
            object will be used to automatically assign a name.
        params (params): Parameters to be passed automatically to
            the tool. This currently allows only for mapping metadata
            fields from the input / task state onto parameters. These
            models precede other parameters that are used by the
            model.
            For example:

            ```python
            @tool(params = dict(color = "metadata.color"))
            def mytool():
                async def execute(color: str, cut: str):
                    ...

                return execute

            ```

    Returns:
        Tool with registry attributes.
    """
    # remove spurous spacing from prompt (can occur if a multline string
    # is used to specify the prompt)
    if prompt:
        prompt = re.sub(r"\s+", " ", prompt)

    def wrapper(tool_type: ToolType) -> ToolType:
        # determine the name (explicit or implicit from object)
        tool_name = registry_name(
            tool_type, name if name else getattr(tool_type, "__name__")
        )

        # wrap instatiations of scorer so they carry registry info and metrics
        def tool_wrapper(*args: Any, **kwargs: Any) -> Tool:
            tool = tool_type(*args, **kwargs)
            registry_tag(
                tool_type,
                tool,
                RegistryInfo(
                    type="tool",
                    name=tool_name,
                    metadata={TOOL_PROMPT: prompt, TOOL_PARAMS: params},
                ),
                *args,
                **kwargs,
            )
            return tool

        # register the scorer
        return tool_register(cast(ToolType, tool_wrapper), tool_name)

    return wrapper


TOOL_PROMPT = "prompt"
TOOL_PARAMS = "params"



import click
from typing_extensions import Unpack

from inspect_ai import eval
from inspect_ai._util.constants import DEFAULT_EPOCHS, DEFAULT_MAX_RETRIES
from inspect_ai._util.samples import parse_samples_limit
from inspect_ai.model import GenerateConfigArgs

from .common import CommonOptions, common_options, resolve_common_options
from .util import parse_cli_args


@click.command("eval")
@click.argument("tasks", nargs=-1)
@click.option(
    "--model",
    type=str,
    required=True,
    envvar=["INSPECT_EVAL_MODEL", "INSPECT_MODEL_NAME"],
    help="Model used to evaluate tasks.",
)
@click.option(
    "--model-base-url",
    type=str,
    help="Base URL for for model API",
)
@click.option(
    "-M",
    multiple=True,
    type=str,
    envvar=["INSPECT_EVAL_MODEL_ARGS"],
    help="One or more native model arguments (e.g. -M arg=value)",
)
@click.option(
    "-T",
    multiple=True,
    type=str,
    envvar="INSPECT_EVAL_TASK_ARGS",
    help="One or more task arguments (e.g. -T arg=value)",
)
@click.option(
    "--limit",
    type=str,
    help="Limit samples to evaluate e.g. 10 or 10,20",
)
@click.option(
    "--epochs",
    type=int,
    help=f"Number of times to repeat dataset (defaults to {DEFAULT_EPOCHS}) ",
)
@click.option(
    "--max-connections",
    type=int,
    help="Maximum number of concurrent connections to Model API (default is per Model API)",
)
@click.option(
    "--max-retries",
    type=int,
    help=f"Maximum number of times to retry request (defaults to {DEFAULT_MAX_RETRIES})",
)
@click.option(
    "--timeout",
    type=int,
    help="Request timeout (in seconds).",
)
@click.option(
    "--max-subprocesses",
    type=int,
    help="Maximum number of subprocesses to run in parallel (default is os.cpu_count())",
)
@click.option(
    "--max-messages",
    type=int,
    help="Maximum number of messages to allow in a task conversation.",
)
@click.option(
    "--no-log-samples",
    type=bool,
    is_flag=True,
    help="Do not include samples in the log file.",
)
@click.option(
    "--no-log-images",
    type=bool,
    is_flag=True,
    help="Do not include base64 encoded versions of filename or URL based images in the log file.",
)
@click.option(
    "--no-score",
    type=bool,
    is_flag=True,
    help="Do not score model output (use the inspect score command to score output later)",
)
@click.option(
    "--max-tokens",
    type=int,
    help="The maximum number of tokens that can be generated in the completion (default is model specific)",
)
@click.option(
    "--system-message",
    type=str,
    help="Override the default system message.",
)
@click.option(
    "--best-of",
    type=int,
    help="Generates best_of completions server-side and returns the 'best' (the one withthe highest log probability per token). OpenAI only.",
)
@click.option(
    "--frequency-penalty",
    type=float,
    help="Number between -2.0 and 2.0. Positive values penalize new tokens based on their existing frequency in the text so far, decreasing the model's likelihood to repeat the same line verbatim. OpenAI only.",
)
@click.option(
    "--presence-penalty",
    type=float,
    help="Number between -2.0 and 2.0. Positive values penalize new tokens based on whether they appear in the text so far, increasing the model's likelihood to talk about new topics. OpenAI only.",
)
@click.option(
    "--logit-bias",
    type=str,
    help='Map token Ids to an associated bias value from -100 to 100 (e.g. "42=10,43=-10")',
)
@click.option("--seed", type=int, help="Random seed. OpenAI only.")
@click.option(
    "--stop-seqs",
    type=str,
    help="Sequences where the API will stop generating further tokens. The returned text will not contain the stop sequence.",
)
@click.option(
    "--suffix",
    type=str,
    help="The suffix that comes after a completion of inserted text. OpenAI only.",
)
@click.option(
    "--temperature",
    type=float,
    help="What sampling temperature to use, between 0 and 2. Higher values like 0.8 will make the output more random, while lower values like 0.2 will make it more focused and deterministic.",
)
@click.option(
    "--top-p",
    type=float,
    help="An alternative to sampling with temperature, called nucleus sampling, where the model considers the results of the tokens with top_p probability mass.",
)
@click.option(
    "--top-k",
    type=int,
    help="Randomly sample the next word from the top_k most likely next words. GDM only.",
)
@click.option(
    "--num-choices",
    type=int,
    help="How many chat completion choices to generate for each input message.",
)
@click.option(
    "--logprobs",
    type=bool,
    is_flag=True,
    help="Return log probabilities of the output tokens. OpenAI and TogetherAI only.",
)
@click.option(
    "--top-logprobs",
    type=int,
    help="Number of most likely tokens (0-20) to return at each token position, each with an associated log probability. OpenAI only.",
)
@common_options
def eval_command(
    tasks: tuple[str] | None,
    model: str,
    model_base_url: str | None,
    m: tuple[str] | None,
    t: tuple[str] | None,
    epochs: int | None,
    limit: str | None,
    max_retries: int | None,
    timeout: int | None,
    max_connections: int | None,
    max_tokens: int | None,
    system_message: str | None,
    best_of: int | None,
    frequency_penalty: float | None,
    presence_penalty: float | None,
    logit_bias: str | None,
    seed: int | None,
    stop_seqs: str | None,
    suffix: str | None,
    temperature: float | None,
    top_p: float | None,
    top_k: int | None,
    num_choices: int | None,
    logprobs: bool | None,
    top_logprobs: int | None,
    max_messages: int | None,
    max_subprocesses: int | None,
    no_log_samples: bool | None,
    no_log_images: bool | None,
    no_score: bool | None,
    **kwargs: Unpack[CommonOptions],
) -> None:
    """Evaluate one or more tasks."""
    # build generate config
    config_keys = list(GenerateConfigArgs.__mutable_keys__)  # type: ignore
    config = GenerateConfigArgs()
    for key, value in locals().items():
        if key in config_keys and value is not None:
            if key == "stop_seqs":
                value = value.split(",")
            if key == "logprobs" and value is False:
                value = None
            config[key] = value  # type: ignore
    # resolve common options
    (log_dir, log_level) = resolve_common_options(kwargs)

    # parse params and model args
    task_args = parse_cli_args(t)
    model_args = parse_cli_args(m)

    # resolve range
    eval_limit = parse_samples_limit(limit)

    # resolve logit_bias
    config["logit_bias"] = parse_logit_bias(logit_bias)

    # resolve negating options
    log_samples = False if no_log_samples else None
    log_images = False if no_log_images else None
    score = False if no_score else True

    # evaluate
    eval(
        tasks=list(tasks) if tasks else None,
        model=model,
        model_base_url=model_base_url,
        model_args=model_args,
        task_args=task_args,
        log_level=log_level,
        log_dir=log_dir,
        limit=eval_limit,
        epochs=epochs,
        max_messages=max_messages,
        max_subprocesses=max_subprocesses,
        log_samples=log_samples,
        log_images=log_images,
        score=score,
        **config,
    )


def parse_logit_bias(logit_bias: str | None) -> dict[int, float] | None:
    logit_biases = parse_cli_args(logit_bias.split(",")) if logit_bias else None
    if logit_biases:
        return dict(
            zip([int(key) for key in logit_biases.keys()], logit_biases.values())
        )
    else:
        return None



import contextlib
import os
from pathlib import Path
from typing import Any, Generator
from urllib.parse import urlparse

from dotenv import dotenv_values, find_dotenv, load_dotenv

from .platform import is_running_in_vscode

INSPECT_LOG_DIR_VAR = "INSPECT_LOG_DIR"


def init_dotenv() -> None:

    # if we are running in vscode, the vscode python extension is already reading in the
    # .env file. This means that editing the .env file within a given session does not
    # actually work! (since load_dotenv doesn't overwrite existing vars by default).
    # so, in this case we actually specify override so we get the more intuitive behavior
    override = is_running_in_vscode()

    # look up the directory tree for a .env file
    dotenv_file = find_dotenv(usecwd=True)

    # we found one, process it
    if dotenv_file:

        # is there an INSPECT_LOG_DIR currently in the environment? (we will give it preference)
        environment_log_dir = os.environ.get(INSPECT_LOG_DIR_VAR, None)
        if environment_log_dir:
            # check for a relative dir, if we find one then resolve to absolute
            fs_scheme = urlparse(environment_log_dir).scheme
            if not fs_scheme and not os.path.isabs(environment_log_dir):
                environment_log_dir = Path(environment_log_dir).resolve().as_posix()

        # is there an INSPECT_LOG_DIR in the .env? If so resolve path relative to .env
        dotenv_log_dir = dotenv_values(dotenv_file).get(INSPECT_LOG_DIR_VAR, None)
        if dotenv_log_dir:
            # check for a relative dir, if we find one then resolve to absolute
            fs_scheme = urlparse(dotenv_log_dir).scheme
            if not fs_scheme and not os.path.isabs(dotenv_log_dir):
                dotenv_log_dir = (
                    (Path(dotenv_file).parent / dotenv_log_dir).resolve().as_posix()
                )

        # do the load, overriding as necessary if we are in vscode
        load_dotenv(dotenv_file, override=override)

        # apply the log_dir, giving preference to the existing environment var
        if environment_log_dir:
            os.environ[INSPECT_LOG_DIR_VAR] = environment_log_dir
        elif dotenv_log_dir:
            os.environ[INSPECT_LOG_DIR_VAR] = dotenv_log_dir


@contextlib.contextmanager
def dotenv_environ(
    override: bool = is_running_in_vscode(),
) -> Generator[Any, Any, None]:
    # determine values to update
    update: dict[str, str] = {}
    values = dotenv_values(".env")
    for key, value in values.items():
        if value is not None and (override or (key not in os.environ.keys())):
            update[key] = value

    # vars to restore and remove on exit
    stomped = set(update.keys()) & set(os.environ.keys())
    update_after = {k: os.environ[k] for k in stomped}
    remove_after = frozenset(k for k in update if k not in os.environ)

    # do the thing
    try:
        os.environ.update(update)
        yield
    finally:
        os.environ.update(update_after)
        [os.environ.pop(k) for k in remove_after]



import asyncio
import contextlib
import datetime
from dataclasses import dataclass
from types import TracebackType
from typing import Any, Callable, Iterator, Type

from rich.align import Align
from rich.console import Console, RenderableType
from rich.live import Live
from rich.panel import Panel
from rich.progress import (
    BarColumn,
    SpinnerColumn,
    TaskProgressColumn,
    TimeElapsedColumn,
)
from rich.progress import Progress as RProgress
from rich.table import Table
from rich.text import Text
from typing_extensions import override

from inspect_ai._util.platform import is_running_in_jupyterlab, is_running_in_vscode
from inspect_ai.log import EvalError, EvalResults, EvalStats
from inspect_ai.log._log import rich_traceback
from inspect_ai.util._context.concurrency import concurrency_status
from inspect_ai.util._context.logger import logger_http_rate_limit_count

from ._display import Display, Progress, TaskDisplay, TaskProfile


@dataclass
class Theme:
    meta: str = "blue"
    light: str = "bright_black"
    metric: str = "green"
    link: str = "blue"


class RichDisplay(Display):
    def __init__(self) -> None:
        self.console = rich_console()
        self.theme = Theme()

    @override
    def print(self, message: str) -> None:
        self.console.print(message, markup=False, highlight=False)

    @override
    @contextlib.contextmanager
    def progress(self, total: int) -> Iterator[Progress]:
        with rich_progress(self.console) as progress:
            yield RichProgress(total, progress)

    @override
    @contextlib.contextmanager
    def task(self, profile: TaskProfile) -> Iterator[TaskDisplay]:
        with Live(None, console=self.console) as live:
            # create task display
            display = RichTaskDisplay(
                profile,
                self.console,
                self.theme,
                lambda r: live.update(r, refresh=True),
            )

            # setup some timed updates (for when no progress ticks are occurring)
            loop = asyncio.get_event_loop()
            handle: asyncio.TimerHandle | None

            def update_display() -> None:
                display.on_update()
                nonlocal handle
                handle = loop.call_later(5, update_display)

            handle = loop.call_later(5, update_display)

            # yield the display
            yield display

            # cleanup handle if we need to
            if handle:
                handle.cancel()


# Note that use of rich progress seems to result in an extra
# empty cell after execution, see:
# https://github.com/Textualize/rich/issues/3211
# https://github.com/Textualize/rich/issues/3168


class RichProgress(Progress):
    def __init__(
        self,
        total: int,
        progress: RProgress,
        on_update: Callable[[], None] | None = None,
    ) -> None:
        self.total = total
        self.progress = progress
        self.task_id = progress.add_task("", total=102)
        self.on_update = on_update

    @override
    def update(self, n: float = 1) -> None:
        advance = (n / self.total) * 100
        self.progress.update(task_id=self.task_id, advance=advance, refresh=True)
        if self.on_update:
            self.on_update()


class RichTaskDisplay(TaskDisplay):
    def __init__(
        self,
        profile: TaskProfile,
        console: Console,
        theme: Theme,
        render: Callable[[RenderableType], None],
    ) -> None:
        self.profile = profile
        self.console = console
        self.theme = theme
        self.progress_ui = rich_progress(console)
        self.render = render
        self.on_update()

    @override
    @contextlib.contextmanager
    def progress(self, total: int) -> Iterator[Progress]:
        yield RichProgress(total, self.progress_ui, self.on_update)

    @override
    def summary(self, results: EvalResults, stats: EvalStats) -> None:
        panel = self.task_panel(
            body=task_stats(self.profile, stats, self.theme),
            config=None,
            footer=task_results(results, self.theme),
            log_location=self.profile.log_location,
        )
        self.render(panel)

    @override
    def error(
        self,
        error: EvalError,
        exc_type: Type[Any],
        exc_value: BaseException,
        traceback: TracebackType | None,
    ) -> None:
        panel = self.task_panel(
            body=rich_traceback(exc_type, exc_value, traceback),
            config=None,
            footer=None,
            log_location=self.profile.log_location,
        )
        self.render(panel)

    def on_update(self) -> None:
        panel = self.task_panel(
            body=Align(self.progress_ui, vertical="middle"),
            config=task_config(self.profile, self.theme),
            footer=live_task_footer(self.theme),
            log_location=None,
        )
        self.render(panel)

    def task_panel(
        self,
        body: RenderableType,
        config: str | None,
        footer: tuple[RenderableType, RenderableType] | None,
        log_location: str | None,
    ) -> Panel:
        return task_panel(
            profile=self.profile,
            body=body,
            config=config,
            footer=footer,
            log_location=log_location,
            options=TaskPanelOptions(
                theme=self.theme,
                # rich doesn't detect vs code width properly
                width=(80 if is_vscode_notebook(self.console) else None),
                jupyter=self.console.is_jupyter,
            ),
        )


@dataclass
class TaskPanelOptions:
    theme: Theme
    width: int | None
    jupyter: bool


def task_panel(
    profile: TaskProfile,
    body: RenderableType,
    config: str | None,
    footer: tuple[RenderableType, RenderableType] | None,
    log_location: str | None,
    options: TaskPanelOptions,
) -> Panel:
    # alias theme
    theme = options.theme

    # setup table
    table = Table.grid(expand=True)
    table.add_column()
    table.add_column(justify="right")

    # main progress and task info
    table.add_row(
        body,
        Text(task_targets(profile), style=theme.meta),
    )

    # config
    if config:
        table.add_row(config)

    # footer if sepecified
    if footer:
        table.add_row()
        table.add_row(footer[0], footer[1])

    # enclose in outer table for log link footer
    root = table
    if log_location:
        # if we are in jupyter then use a real hyperink
        if options.jupyter:
            log_location = f"[link={log_location}]{log_location}[/link]"

        root = Table.grid(expand=True)
        root.add_column()
        root.add_row(table)
        root.add_row()
        root.add_row(
            f"[bold][{theme.light}]Log:[/{theme.light}][/bold] "
            + f"[{theme.link}]{log_location}[/{theme.link}]"
        )

    # create panel w/ title
    panel = Panel(
        root,
        title=f"[bold][{theme.meta}]{task_title(profile)}[/{theme.meta}][/bold]",
        title_align="left",
        width=options.width,
        expand=True,
    )
    return panel


def task_title(profile: TaskProfile) -> str:
    sequence = (
        f"task {profile.sequence[0]}/{profile.sequence[1]}: "
        if profile.sequence[1] > 1
        else ""
    )
    eval_epochs = profile.eval_config.epochs or 1
    epochs = f" x {profile.eval_config.epochs}" if eval_epochs > 1 else ""
    samples = f"{profile.samples//eval_epochs:,}{epochs} sample{'s' if profile.samples > 1 else ''}"
    title = f"{sequence}{profile.name} ({samples})"
    return title


def task_targets(profile: TaskProfile) -> str:
    return "   " + "\n   ".join(
        [str(profile.model), f"dataset: {profile.dataset}", f"scorer: {profile.scorer}"]
    )


def task_config(profile: TaskProfile, theme: Theme) -> str:
    # merge config
    config = (
        dict(profile.task_args)
        | dict(profile.eval_config.model_dump(exclude_none=True))
        | dict(profile.generate_config.model_dump(exclude_none=True))
    )
    config_print: list[str] = []
    for name, value in config.items():
        if name not in ["limit", "epochs"]:
            config_print.append(f"{name}: {value}")
    values = ", ".join(config_print)
    if values:
        return f"[{theme.light}]{values}[/{theme.light}]"
    else:
        return ""


def task_resources() -> str:
    resources: dict[str, str] = {}
    for model, resource in concurrency_status().items():
        resources[model] = f"{resource[0]}/{resource[1]}"
    return task_dict(resources)


def live_task_footer(theme: Theme) -> tuple[RenderableType, RenderableType]:
    return (
        f"[{theme.light}]{task_resources()}[/{theme.light}]",
        Text(task_http_rate_limits(), style=theme.light),
    )


def task_results(
    results: EvalResults, theme: Theme
) -> tuple[RenderableType, RenderableType]:
    output: dict[str, str] = {}
    for name, metric in results.metrics.items():
        value = (
            "1.0"
            if metric.value == 1
            else (
                str(metric.value)
                if isinstance(metric.value, int)
                else f"{metric.value:.3g}"
            )
        )
        output[name] = value
    metrics = f"[{theme.metric}]{task_dict(output, True)}[/{theme.metric}]"

    return (metrics, "")


def task_stats(profile: TaskProfile, stats: EvalStats, theme: Theme) -> RenderableType:
    panel = Table.grid(expand=True)
    panel.add_column()
    config = task_config(profile, theme)
    if config:
        panel.add_row(config)
        panel.add_row()
    elif len(stats.model_usage) < 2:
        panel.add_row()

    table = Table.grid(expand=True)
    table.add_column(style="bold")
    table.add_column()

    # eval time
    started = datetime.datetime.fromisoformat(stats.started_at)
    completed = datetime.datetime.fromisoformat(stats.completed_at)
    elapsed = completed - started
    table.add_row(Text("total time:", style="bold"), f"  {elapsed}", style=theme.light)

    # token usage
    for model, usage in stats.model_usage.items():
        table.add_row(
            Text(model, style="bold"),
            f"  {usage.total_tokens:,} tokens [{usage.input_tokens:,} + {usage.output_tokens:,}]",
            style=theme.light,
        )

    panel.add_row(table)
    return panel


def task_http_rate_limits() -> str:
    return f"HTTP rate limits: {logger_http_rate_limit_count():,}"


def task_dict(d: dict[str, str], bold_value: bool = False) -> str:
    slot1, slot2 = ("", "[/bold]") if bold_value else ("[/bold]", "")
    return "  ".join(
        [f"[bold]{key}:{slot1} {value}{slot2}" for key, value in d.items()]
    )


def rich_progress(console: Console) -> RProgress:
    return RProgress(
        SpinnerColumn(finished_text="✓"),
        BarColumn(bar_width=40 if is_vscode_notebook(console) else None),
        TaskProgressColumn(),
        TimeElapsedColumn(),
        transient=True,
        console=console,
        expand=not is_vscode_notebook(console),
    )


def is_vscode_notebook(console: Console) -> bool:
    return console.is_jupyter and is_running_in_vscode()


def rich_console() -> Console:
    global _console
    if _console is None:
        # only use color in vscode (other terminals are too
        # variable in their color contrast levels to rely on)
        use_color = is_running_in_vscode() and not is_running_in_jupyterlab()
        _console = Console(no_color=not use_color)
    return _console


def rich_display() -> RichDisplay:
    global _display
    if _display is None:
        _display = RichDisplay()
    return _display


_console: Console | None = None
_display: RichDisplay | None = None



import json
import os
import re
from pathlib import Path
from typing import Any, Callable, cast
from urllib.parse import urlparse

import json_stream  # type: ignore
from pydantic import BaseModel, Field
from pydantic_core import to_json

from inspect_ai._util.file import FileInfo, file, filesystem

from ._log import (
    EvalError,
    EvalLog,
    EvalPlan,
    EvalResults,
    EvalSample,
    EvalSpec,
    EvalStats,
    LogEvent,
    LoggingMessage,
    Recorder,
)


class EvalLogInfo(FileInfo):
    task: str
    """Task name."""

    task_id: str
    """Task id."""

    suffix: str | None
    """Log file suffix (e.g. "-scored")"""


def list_eval_logs(
    log_dir: str = os.environ.get("INSPECT_LOG_DIR", "./logs"),
    filter: Callable[[EvalLog], bool] | None = None,
    recursive: bool = True,
    extensions: list[str] = [".json", ".jsonl"],
    descending: bool = True,
    fs_options: dict[str, Any] = {},
) -> list[EvalLogInfo]:
    """List all eval logs in a directory.

    Args:
      log_dir (str): Log directory (defaults to INSPECT_LOG_DIR)
      filter (Callable[[EvalLog], bool]): Filter to limit logs returned.
         Note that the EvalLog instance passed to the filter has only
         the EvalLog header (i.e. does not have the samples or logging output).
      recursive (bool): List log files recursively (defaults to True).

      extensions (list[str]): File extension to scan for logs
      descending (bool): List in descening order.
      fs_options (dict[str, Any]): Optional. Addional arguments to pass through
          to the filesystem provider (e.g. `S3FileSystem`).

    Returns:
       List of EvalLog Info.

    """
    # get the eval logs
    fs = filesystem(log_dir, fs_options)
    if fs.exists(log_dir):
        eval_logs = log_files_from_ls(
            fs.ls(log_dir, recursive=recursive), extensions, descending
        )
    else:
        return []

    # apply filter if requested
    if filter:
        return [
            log
            for log in eval_logs
            if filter(read_eval_log(log.name, header_only=True))
        ]
    else:
        return eval_logs


def write_eval_log(log: EvalLog, log_file: str | FileInfo) -> None:
    """Write an evaluation log.

    Args:
       log (EvalLog): Evaluation log to write.
       log_file (str | FileInfo): Location to write log to.

    """
    log_file = log_file if isinstance(log_file, str) else log_file.name
    with file(log_file, "w") as f:
        f.write(eval_log_json(log))


def eval_log_json(log: EvalLog) -> str:
    # serialize to json (ignore values that are unserializable)
    # these values often result from solvers using metadata to
    # pass around 'live' objects -- this is fine to do and we
    # don't want to prevent it at the serialization level
    return to_json(
        value=log, indent=2, exclude_none=True, fallback=lambda _x: None
    ).decode()


def read_eval_log(log_file: str | FileInfo, header_only: bool = False) -> EvalLog:
    """Read an evaluation log.

    Args:
       log_file (str | FileInfo): Log file to read.
       header_only (bool): Read only the header (i.e. exclude
         the "samples" and "logging" fields). Defaults to False.

    Returns:
       EvalLog object read from file.
    """
    log_file = log_file if isinstance(log_file, str) else log_file.name
    with file(log_file, "r") as f:
        # header-only uses json-stream
        if header_only:
            data = json_stream.load(f, persistent=True)

            def read_field(field: str) -> Any:
                if field in data.keys():
                    return json_stream.to_standard_types(data[field])
                else:
                    return None

            results = read_field("results")
            error = read_field("error")

            return EvalLog(
                version=read_field("version"),
                status=read_field("status"),
                eval=EvalSpec(**read_field("eval")),
                plan=EvalPlan(**read_field("plan")),
                results=EvalResults(**results) if results else None,
                stats=EvalStats(**read_field("stats")),
                error=EvalError(**error) if error else None,
            )

        # otherwise normal json parse
        else:
            raw_data = json.load(f)
            log = EvalLog(**raw_data)
            if log.version > 1:
                raise ValueError(f"Unable to read version {log.version} of log format.")
            return log


class FileRecorder(Recorder):
    def __init__(
        self, log_dir: str, suffix: str, fs_options: dict[str, Any] = {}
    ) -> None:
        super().__init__()
        self.log_dir = log_dir
        self.fs = filesystem(log_dir, fs_options)
        self.fs.mkdir(self.log_dir, exist_ok=True)
        self.suffix = suffix

    def latest_log_file_path(self) -> str:
        log_files = self.fs.ls(self.log_dir)
        sorted_log_files = log_files_from_ls(log_files, [self.suffix])
        if len(sorted_log_files) > 0:
            log_file = sorted_log_files[0].name
            # return as relative if the fs_scheme is a local relative path
            fs_scheme = urlparse(self.log_dir).scheme
            if not fs_scheme and not os.path.isabs(self.log_dir):
                log_dir_abs = Path(self.log_dir).parent.absolute().as_uri()
                log_file = log_file.replace(log_dir_abs, ".")
            return log_file
        else:
            raise FileNotFoundError("No evaluation logs found in in output_dir")

    def _log_file_key(self, eval: EvalSpec) -> str:
        # clean underscores, slashes, and : from the log file key (so we can reliably parse it
        # later without worrying about underscores)
        def clean(s: str) -> str:
            return s.replace("_", "-").replace("/", "-").replace(":", "-")

        return f"{clean(eval.created)}_{clean(eval.task)}_{clean(eval.task_id)}"

    def _log_file_path(self, eval: EvalSpec) -> str:
        return f"{self.log_dir}{self.fs.sep}{self._log_file_key(eval)}{self.suffix}"


def log_files_from_ls(
    ls: list[FileInfo],
    extensions: list[str] = [".json", ".jsonl"],
    descending: bool = True,
) -> list[EvalLogInfo]:
    return [
        log_file_info(file)
        for file in sorted(ls, key=lambda file: file.mtime, reverse=descending)
        if file.type == "file" and is_log_file(file.name, extensions)
    ]


log_file_pattern = r"^\d{4}-\d{2}-\d{2}T\d{2}[:-]\d{2}[:-]\d{2}.*$"


def is_log_file(file: str, extensions: list[str]) -> bool:
    parts = file.replace("\\", "/").split("/")
    name = parts[-1]
    return re.match(log_file_pattern, name) is not None and any(
        [name.endswith(suffix) for suffix in extensions]
    )


def log_file_info(info: FileInfo) -> "EvalLogInfo":
    # extract the basename and split into parts
    # (deal with previous logs had the model in their name)
    basename = os.path.splitext(info.name)[0]
    parts = basename.split("/").pop().split("_")
    last_idx = 3 if len(parts) > 3 else 2
    task = parts[1]
    part3 = parts[last_idx].split("-")
    task_id = part3[0]
    suffix = task_id[2] if len(part3) > 1 else None
    return EvalLogInfo(
        name=info.name,
        type=info.type,
        size=info.size,
        mtime=info.mtime,
        task=task,
        task_id=task_id,
        suffix=suffix,
    )


class JSONRecorder(FileRecorder):
    class JSONLogFile(BaseModel):
        file: str
        data: EvalLog
        events: int = Field(default=0)

    def __init__(self, log_dir: str, write_freq: int = 100):
        # call super
        super().__init__(log_dir, ".json")

        # flush to file every write_freq events
        self.write_freq = write_freq

        # each eval has a unique key (created from run_id and task name/version)
        # which we use to track the output path, accumulated data, and event counter
        self.data: dict[str, JSONRecorder.JSONLogFile] = {}

    def log_start(self, eval: EvalSpec) -> str:
        # initialize file log for this eval
        file = self._log_file_path(eval)
        self.data[self._log_file_key(eval)] = JSONRecorder.JSONLogFile(
            file=file,
            data=EvalLog(eval=eval),
            events=0,
        )
        return file

    def log_event(
        self,
        spec: EvalSpec,
        type: LogEvent,
        data: EvalPlan | EvalSample | EvalResults | LoggingMessage,
    ) -> None:
        log = self.data[self._log_file_key(spec)]
        if type == "plan":
            log.data.plan = cast(EvalPlan, data)
        elif type == "sample":
            if log.data.samples is None:
                log.data.samples = []
            log.data.samples.append(cast(EvalSample, data))
        elif type == "logging":
            log.data.logging.append(cast(LoggingMessage, data))
        elif type == "results":
            log.data.results = cast(EvalResults, data)
        else:
            raise ValueError(f"Unknown event {type}")
        # check if we need to flush
        if log.events >= self.write_freq:
            self.write_log(log.file, log.data)
            log.events = 0
        log.events += 1

    def log_success(
        self,
        spec: EvalSpec,
        stats: EvalStats,
    ) -> EvalLog:
        log = self.data[self._log_file_key(spec)]
        log.data.status = "success"
        log.data.stats = stats
        return self._log_finish(spec, log)

    def log_failure(
        self, spec: EvalSpec, stats: EvalStats, error: EvalError
    ) -> EvalLog:
        log = self.data[self._log_file_key(spec)]
        log.data.status = "error"
        log.data.stats = stats
        log.data.error = error
        return self._log_finish(spec, log)

    def read_log(self, location: str) -> EvalLog:
        return read_eval_log(location)

    def write_log(self, location: str, log: EvalLog) -> None:
        write_eval_log(log, location)

    def read_latest_log(self) -> EvalLog:
        return self.read_log(self.latest_log_file_path())

    def _log_finish(self, spec: EvalSpec, log: JSONLogFile) -> EvalLog:
        self.write_log(log.file, log.data)
        del self.data[self._log_file_key(spec)]
        return log.data



import json
import os
from typing import Any

from mistralai.async_client import MistralAsyncClient
from mistralai.exceptions import MistralAPIStatusException
from mistralai.models.chat_completion import (
    ChatCompletionResponse,
    ChatCompletionResponseChoice,
    FinishReason,
    FunctionCall,
    ToolType,
)
from mistralai.models.chat_completion import (
    ChatMessage as MistralChatMessage,
)
from mistralai.models.chat_completion import (
    ToolCall as MistralToolCall,
)
from mistralai.models.chat_completion import (
    ToolChoice as MistralToolChoice,
)
from typing_extensions import override

from inspect_ai._util.constants import (
    DEFAULT_MAX_RETRIES,
    DEFAULT_MAX_TOKENS,
    DEFAULT_TIMEOUT,
)
from inspect_ai.model._providers.util import model_base_url

from .._model import (
    ChatCompletionChoice,
    ChatMessage,
    ChatMessageAssistant,
    GenerateConfig,
    ModelAPI,
    ModelOutput,
    ModelUsage,
    StopReason,
)
from .._tool import ToolCall, ToolChoice, ToolFunction, ToolInfo
from .._util import chat_api_tool

AZURE_MISTRAL_API_KEY = "AZURE_MISTRAL_API_KEY"
AZUREAI_MISTRAL_API_KEY = "AZUREAI_MISTRAL_API_KEY"
MISTRAL_API_KEY = "MISTRAL_API_KEY"


class MistralAPI(ModelAPI):
    def __init__(
        self,
        model_name: str,
        base_url: str | None = None,
        config: GenerateConfig = GenerateConfig(),
        **model_args: Any,
    ):
        super().__init__(model_name=model_name, base_url=base_url, config=config)

        # resolve api_key -- look for mistral then azure
        api_key = os.environ.get(MISTRAL_API_KEY, None)
        if api_key:
            base_url = model_base_url(base_url, "MISTRAL_BASE_URL")
            if base_url:
                model_args["endpoint"] = base_url
        else:
            api_key = os.environ.get(
                AZUREAI_MISTRAL_API_KEY, os.environ.get(AZURE_MISTRAL_API_KEY, None)
            )
            if not api_key:
                raise ValueError(
                    f"{MISTRAL_API_KEY} or {AZUREAI_MISTRAL_API_KEY} environment variable not found."
                )
            base_url = model_base_url(base_url, "AZUREAI_MISTRAL_BASE_URL")
            if not base_url:
                raise ValueError(
                    "You must provide a base URL when using Mistral on Azure. Use the AZUREAI_MISTRAL_BASE_URL "
                    + " environment variable or the --model-base-url CLI flag to set the base URL."
                )
            model_args["endpoint"] = base_url

        # save key
        self.api_key = api_key

        # create client
        self.client = MistralAsyncClient(
            api_key=api_key,
            max_retries=(
                config.max_retries if config.max_retries else DEFAULT_MAX_RETRIES
            ),
            timeout=config.timeout if config.timeout else DEFAULT_TIMEOUT,
            **model_args,
        )

    async def generate(
        self,
        input: list[ChatMessage],
        tools: list[ToolInfo],
        tool_choice: ToolChoice,
        config: GenerateConfig,
    ) -> ModelOutput:
        # send request
        response = await self.client.chat(
            model=self.model_name,
            messages=[mistral_chat_message(message) for message in input],
            temperature=config.temperature,
            top_p=config.top_p,
            max_tokens=config.max_tokens,
            random_seed=config.seed,
            tools=mistral_chat_tools(tools) if len(tools) > 0 else None,
            tool_choice=(
                mistral_chat_tool_choice(tool_choice) if len(tools) > 0 else None
            ),
        )

        # return model output (w/ tool calls if they exist)
        choices = completion_choices_from_response(response)
        return ModelOutput(
            model=response.model,
            choices=choices,
            usage=ModelUsage(
                input_tokens=response.usage.prompt_tokens,
                output_tokens=(
                    response.usage.completion_tokens
                    if response.usage.completion_tokens
                    else response.usage.total_tokens - response.usage.prompt_tokens
                ),
                total_tokens=response.usage.total_tokens,
            ),
        )

    @override
    def is_rate_limit(self, ex: BaseException) -> bool:
        return isinstance(ex, MistralAPIStatusException) and ex.http_status == 429

    @override
    def connection_key(self) -> str:
        return self.api_key

    # not clear what the mistral default max tokens is (not documented)
    # so we set it to the default to be sure
    @override
    def max_tokens(self) -> int:
        return DEFAULT_MAX_TOKENS


def mistral_chat_tools(tools: list[ToolInfo]) -> list[dict[str, Any]]:
    chat_tools = [chat_api_tool(tool) for tool in tools]
    return [dict(type=tool["type"], function=tool["function"]) for tool in chat_tools]


def mistral_chat_tool_choice(tool_choice: ToolChoice) -> MistralToolChoice:
    if isinstance(tool_choice, ToolFunction):
        # mistral doesn't support specifically named tools to use
        # (rather just 'any' which says use at least one tool)
        return MistralToolChoice.any
    elif tool_choice == "auto":
        return MistralToolChoice.auto
    else:
        return MistralToolChoice.none


def mistral_chat_message(message: ChatMessage) -> MistralChatMessage:
    if message.role == "assistant" and message.tool_calls:
        return MistralChatMessage(
            role=message.role,
            content=message.text,
            tool_calls=[mistral_tool_call(call) for call in message.tool_calls],
        )
    elif message.role == "tool":
        return MistralChatMessage(
            role=message.role,
            name=message.tool_call_id,
            content=(
                f"Error: {message.tool_error}" if message.tool_error else message.text
            ),
        )
    else:
        return MistralChatMessage(role=message.role, content=message.text)


def mistral_tool_call(tool_call: ToolCall) -> MistralToolCall:
    return MistralToolCall(
        id=tool_call.id,
        type=ToolType.function,
        function=mistral_function_call(tool_call),
    )


def mistral_function_call(tool_call: ToolCall) -> FunctionCall:
    return FunctionCall(
        name=tool_call.function, arguments=json.dumps(tool_call.arguments)
    )


def chat_tool_calls(message: MistralChatMessage) -> list[ToolCall] | None:
    if message.tool_calls:
        return [
            ToolCall(
                id=call.id,
                function=call.function.name,
                arguments=json.loads(call.function.arguments),
                type="function",
            )
            for call in message.tool_calls
        ]
    else:
        return None


def completion_choice(choice: ChatCompletionResponseChoice) -> ChatCompletionChoice:
    message = choice.message
    completion = message.content
    if isinstance(completion, list):
        completion = " ".join(completion)
    return ChatCompletionChoice(
        message=ChatMessageAssistant(
            content=completion, tool_calls=chat_tool_calls(message), source="generate"
        ),
        stop_reason=(
            choice_stop_reason(choice)
            if choice.finish_reason is not None
            else "unknown"
        ),
    )


def completion_choices_from_response(
    response: ChatCompletionResponse,
) -> list[ChatCompletionChoice]:
    return [completion_choice(choice) for choice in response.choices]


def choice_stop_reason(choice: ChatCompletionResponseChoice) -> StopReason:
    match choice.finish_reason:
        case FinishReason.stop:
            return "stop"
        case FinishReason.length:
            return "length"
        case FinishReason.tool_calls:
            return "tool_calls"
        case _:
            return "unknown"



import pytest

from inspect_ai.model import (
    ChatMessageAssistant,
    ChatMessageUser,
    ContentImage,
    ContentText,
)
from inspect_ai.model._model import collapse_consecutive_user_messages


@pytest.fixture
def user_message_str():
    return ChatMessageUser(content="User message")


@pytest.fixture
def user_message_image_and_str():
    return ChatMessageUser(
        content=[ContentImage(image="foo"), ContentText(text="Message")]
    )


@pytest.fixture
def assistant_message():
    return ChatMessageAssistant(content="Assistant message")


@pytest.fixture
def combined_user_message():
    return ChatMessageUser(
        content=[ContentText(text="Message 1"), ContentText(text="Message 2")]
    )


def test_collapse_consecutive_user_messages_single_user_message(user_message_str):
    messages = [user_message_str]
    assert collapse_consecutive_user_messages(messages) == messages


def test_collapse_consecutive_user_messages_alternating_messages(
    user_message_str, assistant_message
):
    messages = [user_message_str, assistant_message, user_message_str]
    assert collapse_consecutive_user_messages(messages) == messages


def test_collapse_consecutive_user_messages_consecutive_user_messages(user_message_str):
    messages = [user_message_str, user_message_str, user_message_str]
    assert len(collapse_consecutive_user_messages(messages)) == 1


def test_collapse_consecutive_user_messages_with_image_message(
    user_message_image_and_str,
):
    messages = [user_message_image_and_str, user_message_image_and_str]
    assert len(collapse_consecutive_user_messages(messages)) == 1
    assert isinstance(
        collapse_consecutive_user_messages(messages)[0].content[0], ContentImage
    )



import os
from logging import (
    INFO,
    WARNING,
    LogRecord,
    addLevelName,
    getLevelName,
    getLogger,
)

from rich.console import ConsoleRenderable
from rich.logging import RichHandler
from rich.text import Text
from typing_extensions import override

from inspect_ai._util.constants import (
    DEFAULT_LOG_LEVEL,
    HTTP,
    HTTP_LOG_LEVEL,
    PKG_NAME,
)
from inspect_ai.util._context.logger import notify_logger_record

from .rich import rich_console


# log handler that filters messages to stderr and the log file
class LogHandler(RichHandler):
    def __init__(self, levelno: int) -> None:
        super().__init__(levelno, console=rich_console())
        self.display_level = WARNING

    @override
    def emit(self, record: LogRecord) -> None:
        # demote httpx and retury notifications to log_level http
        if record.name == "httpx" or "Retrying request" in record.getMessage():
            record.levelno = HTTP
            record.levelname = HTTP_LOG_LEVEL

        # skip httpx event loop is closed errors
        if "Event loop is closed" in record.getMessage():
            return

        # write to stderr if we are at or above the threshold
        if record.levelno >= self.display_level:
            super().emit(record)

        # eval log always gets info level and higher records
        # eval log only gets debug or http if we opt-in
        write = record.levelno >= INFO or record.levelno >= self.display_level
        notify_logger_record(record, write)

    @override
    def render_message(self, record: LogRecord, message: str) -> ConsoleRenderable:
        return Text.from_ansi(message)


# initialize logging -- this function can be called multiple times
# in the lifetime of the process (the levelno will update globally)
def init_logger(log_level: str | None = None) -> None:
    # register http level
    addLevelName(HTTP, HTTP_LOG_LEVEL)

    # resolve default log level
    log_level = (
        log_level if log_level else os.getenv("INSPECT_LOG_LEVEL", DEFAULT_LOG_LEVEL)
    )

    # convert to integer
    levelno = getLevelName(log_level.upper())

    # init logging handler on demand
    global _logHandler
    if not _logHandler:
        _logHandler = LogHandler(min(HTTP, levelno))
        getLogger().addHandler(_logHandler)

    # establish default capture level
    capture_level = min(HTTP, levelno)

    # see all the messages (we won't actually display/write all of them)
    getLogger().setLevel(capture_level)
    getLogger(PKG_NAME).setLevel(capture_level)
    getLogger("httpx").setLevel(capture_level)

    # set the levelno on the global handler
    _logHandler.display_level = levelno


_logHandler: LogHandler | None = None



import abc
import contextlib
from dataclasses import dataclass
from types import TracebackType
from typing import Any, Iterator, Type

from inspect_ai.log import EvalConfig, EvalError, EvalResults, EvalStats
from inspect_ai.model import GenerateConfig, ModelName


class Progress(abc.ABC):
    @abc.abstractmethod
    def update(self, n: float = 1) -> None: ...


class TaskDisplay(abc.ABC):
    @abc.abstractmethod
    @contextlib.contextmanager
    def progress(self, total: int) -> Iterator[Progress]: ...

    @abc.abstractmethod
    def summary(self, results: EvalResults, stats: EvalStats) -> None: ...

    @abc.abstractmethod
    def error(
        self,
        error: EvalError,
        exc_type: Type[Any],
        exc_value: BaseException,
        traceback: TracebackType | None,
    ) -> None: ...


@dataclass
class TaskProfile:
    name: str
    sequence: tuple[int, int]
    model: ModelName
    dataset: str
    scorer: str
    samples: int
    eval_config: EvalConfig
    task_args: dict[str, Any]
    generate_config: GenerateConfig
    log_location: str


class Display(abc.ABC):
    @abc.abstractmethod
    def print(self, message: str) -> None: ...

    @abc.abstractmethod
    @contextlib.contextmanager
    def progress(self, total: int) -> Iterator[Progress]: ...

    @abc.abstractmethod
    @contextlib.contextmanager
    def task(self, profile: TaskProfile) -> Iterator[TaskDisplay]: ...



import pytest
from utils import skip_if_no_openai, skip_if_no_together

from inspect_ai.model import GenerateConfig, get_model


async def generate(model_name):
    model = get_model(model_name)
    return await model.generate(input="Hello.", config=GenerateConfig(num_choices=3))


async def check_num_choices(model_name):
    model = get_model(model_name)
    response = await model.generate(
        input="Hello.", config=GenerateConfig(num_choices=3)
    )
    assert len(response.choices) == 3


@pytest.mark.asyncio
@skip_if_no_openai
async def test_openai_num_choices() -> None:
    await check_num_choices("openai/gpt-3.5-turbo")


@pytest.mark.asyncio
@skip_if_no_together
async def test_together_num_choices() -> None:
    await check_num_choices("together/google/gemma-2b-it")


# @pytest.mark.asyncio
# @skip_if_no_azureai
# async def test_azureai_num_choices() -> None:
#     await check_num_choices(None)



import json
import os
import subprocess
from pathlib import Path
from typing import Any

from inspect_ai.log import EvalLog

WWW_DIR = os.path.abspath((Path(__file__).parent / "www").as_posix())


def sync_view_schema() -> None:
    """Genreate a JSON schema and Typescript types for EvalLog.

    This is useful for keeping log file viewer JS development
    in sync w/ Python development
    """
    # export schema file
    schema_path = Path(WWW_DIR, "log-schema.json")
    types_path = Path(WWW_DIR, "log.d.ts")
    with open(schema_path, "w", encoding="utf-8") as f:
        # make everything required
        schema = EvalLog.model_json_schema()
        defs: dict[str, Any] = schema["$defs"]
        for key in defs.keys():
            defs[key] = schema_to_strict(defs[key])
        f.write(json.dumps(schema, indent=2))

        # generate types w/ json-schema-to-typescript
        subprocess.run(
            [
                "json2ts",
                "--input",
                schema_path,
                "--output",
                types_path,
                "--additionalProperties",
                "false",
            ]
        )


def schema_to_strict(schema: dict[str, Any]) -> dict[str, Any]:
    properties = schema.get("properties", None)
    if properties:
        schema["required"] = list(properties.keys())
        schema["additionalProperties"] = False

    return schema


if __name__ == "__main__":
    sync_view_schema()



from pathlib import Path
from typing import Callable

from inspect_ai._eval.list import list_tasks
from inspect_ai._eval.task import TaskInfo

TEST_TASKS_DIR = Path("tests/test_task_list")


def list_test_tasks_dir(
    globs: list[str], filter: Callable[[TaskInfo], bool] | None = None
):
    return list_tasks(globs, filter=filter, root_dir=TEST_TASKS_DIR)


def test_task_list_multiple_file():
    tasks = list_test_tasks_dir(["multiple.py"])
    assert len(tasks) == 2
    names = [task.name for task in tasks]
    assert "first" in names
    assert "second_task" in names


def test_task_list_multiple_dir():
    tasks = list_test_tasks_dir(["multiple_dir"])
    assert len(tasks) == 2


def test_task_list_attribs():
    tasks = list_test_tasks_dir(["attribs.ipynb"])
    assert tasks[0].attribs.get("light") is True
    assert tasks[0].attribs.get("type") == "bio"


def test_task_list_filter():
    tasks = list_test_tasks_dir(["*"], filter=lambda t: t.attribs.get("type") == "bio")
    assert len(tasks) == 1


def test_task_list_recurse():
    tasks = list_test_tasks_dir(["recurse"])
    assert len(tasks) == 3



import datetime
import io
from contextlib import contextmanager
from copy import deepcopy
from typing import Any, BinaryIO, Iterator, Literal, cast, overload
from urllib.parse import urlparse

import fsspec  # type: ignore
from pydantic import BaseModel

# https://filesystem-spec.readthedocs.io/en/latest/_modules/fsspec/spec.html#AbstractFileSystem
# https://filesystem-spec.readthedocs.io/en/latest/api.html#fsspec.generic.GenericFileSystem


OpenTextMode = Literal["r", "a", "w"]
OpenBinaryMode = Literal["rb", "ab", "wb"]


@overload
@contextmanager
def file(
    file: str,
    mode: OpenTextMode,
    compression: str | None = "infer",
    encoding: str = "utf-8",
    fs_options: dict[str, Any] = {},
) -> Iterator[io.TextIOWrapper]:
    ...


@overload
@contextmanager
def file(
    file: str,
    mode: OpenBinaryMode,
    compression: str | None = "infer",
    encoding: str = "utf-8",
    fs_options: dict[str, Any] = {},
) -> Iterator[BinaryIO]:
    ...


@contextmanager
def file(
    file: str,
    mode: OpenTextMode | OpenBinaryMode,
    compression: str | None = "infer",
    encoding: str = "utf-8",
    fs_options: dict[str, Any] = {},
) -> Iterator[io.TextIOWrapper] | Iterator[BinaryIO]:
    open
    """Open local or remote file stream.

    Open a file stream for reading or writing. Refer to a local file or
    use a URI with a remove filesystem prefix (e.g. 's3://'). The
    `fsspec` package is used to resolve filesystem URLs.

    Args:
        file (str):
          Local file path or remove filesystem URL (e.g. 's3://')
        mode (str): Mode for accessing file ("r", "rb", "w", "wb", etc.).
        compression (str | None): Compression used by file. See
          `fsspec.available_compressions()`. Default to "infer",
          which will infer the compression from the file extension.
        encoding: (str): Encoding for text files (defaults to "utf-8")
        fs_options (dict[str, Any]): Optional. Addional arguments to pass through
          to the filesystem provider (e.g. `S3FileSystem`). Use `{"anon": True }`
          if you are accessing a public S3 bucket with no credentials.

    """
    # get the default storage options for the scheme then apply passed options
    options = default_fs_options(file)
    options.update(fs_options)

    # open the file
    open_file = fsspec.open(
        file, mode=mode, compression=compression, encoding=encoding, **options
    )

    # yield the file and ensure it is closed when we exit the context
    with open_file as f:
        try:
            yield f
        finally:
            f.close()


class FileInfo(BaseModel):
    name: str
    """Name of file."""

    type: str
    """Type of file (file or dir)"""

    size: int
    """File size in bytes."""

    mtime: float
    """File modification time."""


class FileSystem:
    def __init__(self, fs: Any) -> None:
        self.fs = fs

    @property
    def sep(self) -> str:
        return cast(str, self.fs.sep)

    def exists(self, path: str) -> bool:
        return self.fs.exists(path) is True

    def mkdir(self, path: str, exist_ok: bool = False) -> None:
        self.fs.makedirs(path, exist_ok=exist_ok)

    def ls(
        self, path: str, recursive: bool = False, **kwargs: dict[str, Any]
    ) -> list[FileInfo]:
        # prevent caching of listings
        self.fs.invalidate_cache(path)

        # enumerate the files
        if recursive:
            files: list[dict[str, Any]] = []
            for _, _, filenames in self.fs.walk(path=path, detail=True, **kwargs):
                files.extend(filenames.values())
        else:
            files = cast(
                list[dict[str, Any]],
                self.fs.ls(path, detail=True, **kwargs),
            )

        # fixup name and discover mtime
        for info in files:
            # name needs the protocol prepended
            info["name"] = self.fs.unstrip_protocol(info["name"])

            # S3 filesystems use "LastModified"
            if "LastModified" in info.keys():
                info["mtime"] = cast(
                    datetime.datetime, cast(Any, info)["LastModified"]
                ).timestamp()
            # if we don't yet have an mtime key then fetch created explicitly
            if "mtime" not in info.keys():
                info["mtime"] = self.fs.created(file).timestamp()
            info["mtime"] = info["mtime"] * 1000

        # convert to FileInfo
        return [
            FileInfo(
                name=file["name"],
                type=file["type"],
                size=file["size"],
                mtime=file["mtime"],
            )
            for file in files
        ]


def filesystem(path: str, fs_options: dict[str, Any] = {}) -> FileSystem:
    """Return the filesystem used to host the specified path.

    Args:
      path (str): Local path or remote URL e.g. s3://).  The
        `fsspec` package is used to resolve filesystem URLs.
      fs_options (dict[str, Any]): Optional. Addional arguments to pass through
        to the filesystem provider (e.g. `S3FileSystem`). Use `{"anon": True }`
        if you are accessing a public S3 bucket with no credentials.

    Returns:
       An tuple with an `fsspec` compatible filesystem and the
       file-systems-specific URL for file.
    """
    # determine options
    options = default_fs_options(path)
    options.update(fs_options)

    # create filesystem
    fs, path = fsspec.core.url_to_fs(path)
    return FileSystem(fs)


def default_fs_options(file: str) -> dict[str, Any]:
    options = deepcopy(DEFAULT_FS_OPTIONS.get(urlparse(file).scheme, {}))
    # disable caching for all filesystems
    options.update(
        dict(
            skip_instance_cache=False,
            use_listings_cache=False,
        )
    )
    return options


DEFAULT_FS_OPTIONS: dict[str, dict[str, Any]] = dict(
    # disable all S3 native caching
    s3=dict(default_fill_cache=False, default_cache_type="none", cache_regions=False)
)



import glob
import json
import os
import posixpath
from http import HTTPStatus
from http.server import SimpleHTTPRequestHandler
from io import BytesIO
from typing import Any
from urllib.parse import parse_qs, urlparse

from .dev import is_dev_mode


class InspectHTTPRequestHandler(SimpleHTTPRequestHandler):
    def __init__(self, *args: Any, directory: str, **kwargs: Any) -> None:
        # note whether we are in dev mode (i.e. developing the package)
        self.dev_mode = is_dev_mode()

        # initialize file serving directory
        directory = os.path.abspath(directory)
        super().__init__(*args, directory=directory, **kwargs)

    def do_GET(self) -> None:
        if self.path.startswith("/api/events"):
            self.handle_events()
        else:
            super().do_GET()

    def handle_events(self) -> None:
        """Client polls for events (e.g. dev reload) ~ every 1 second."""
        query = parse_qs(urlparse(self.path).query)
        params = dict(zip(query.keys(), [value[0] for value in query.values()]))
        self.send_json(json.dumps(self.events_response(params)))

    def events_response(self, params: dict[str, str]) -> list[str]:
        """Send back a 'reload' event if we have modified source files."""
        loaded_time = params.get("loaded_time", None)
        return (
            ["reload"] if loaded_time and self.should_reload(int(loaded_time)) else []
        )

    def translate_path(self, path: str) -> str:
        """Ensure that paths don't escape self.directory."""
        translated = super().translate_path(path)
        if not os.path.abspath(translated).startswith(self.directory):
            return self.directory
        else:
            return translated

    def send_json(self, json: str | bytes) -> None:
        if isinstance(json, str):
            json = json.encode()
        self.send_response(HTTPStatus.OK)
        self.send_header("Content-type", "application/json")
        self.end_headers()
        self.copyfile(BytesIO(json), self.wfile)  # type: ignore

    def send_response(self, code: int, message: str | None = None) -> None:
        """No client side or proxy caches."""
        super().send_response(code, message)
        self.send_header("Expires", "Fri, 01 Jan 1990 00:00:00 GMT")
        self.send_header("Pragma", "no-cache")
        self.send_header(
            "Cache-Control", "no-cache, no-store, max-age=0, must-revalidate"
        )

    def guess_type(self, path: str | os.PathLike[str]) -> str:
        _, ext = posixpath.splitext(path)
        if not ext or ext == ".mjs" or ext == ".js":
            return "application/javascript"
        elif ext == ".md":
            return "text/markdown"
        else:
            return super().guess_type(path)

    def log_error(self, format: str, *args: Any) -> None:
        if self.dev_mode:
            super().log_error(format, *args)

    def log_request(self, code: int | str = "-", size: int | str = "-") -> None:
        """Don't log status 200 or 404 (too chatty)."""
        if code not in [200, 404]:
            super().log_request(code, size)

    def should_reload(self, loaded_time: int) -> bool:
        if self.dev_mode:
            for dir in self.reload_dirs():
                files = [
                    os.stat(file).st_mtime
                    for file in glob.glob(f"{dir}/**/*", recursive=True)
                ]
                last_modified = max(files) * 1000
                if last_modified > loaded_time:
                    return True

        return False

    def reload_dirs(self) -> list[str]:
        return [self.directory]



"""
Think you have Solved Question Answering? Try ARC, the AI2 Reasoning Challenge

Peter Clark, Isaac Cowhey, Oren Etzioni, Tushar Khot, Ashish Sabharwal, Carissa Schoenick, Oyvind Tafjord
https://arxiv.org/abs/1803.05457

# run all subsets
inspect eval arc.py

# run specific subsets
inspect eval arc.py@arc_easy
inspect eval arc.py@arc_challenge
"""

from inspect_ai import Task, task
from inspect_ai.dataset import Sample, hf_dataset
from inspect_ai.scorer import answer
from inspect_ai.solver import multiple_choice


def record_to_sample(record):
    # read the labels and text
    choices = record["choices"]
    choices = dict(zip(choices["label"], choices["text"]))

    # determine the target then normalize to letter
    answerKey = record["answerKey"]
    target = list(choices.keys()).index(answerKey)
    target = chr(ord("A") + int(target))

    # return sample
    return Sample(
        input=record["question"], choices=list(choices.values()), target=target
    )


def arc_task(dataset_name):
    return Task(
        dataset=hf_dataset(
            path="allenai/ai2_arc",
            name=dataset_name,
            split="test",
            sample_fields=record_to_sample,
        ),
        plan=multiple_choice(),
        scorer=answer("letter"),
    )


@task
def arc_easy():
    return arc_task("ARC-Easy")


@task
def arc_challenge():
    return arc_task("ARC-Challenge")



from inspect_ai import Task, task
from inspect_ai.dataset import example_dataset
from inspect_ai.scorer import model_graded_fact
from inspect_ai.solver import chain_of_thought, generate, self_critique


@task
def theory_of_mind(critique = False):

    # use self_critique if requested
    plan = [chain_of_thought(), generate()]
    if critique:
        plan.append(self_critique())

    return Task(
        dataset=example_dataset("theory_of_mind"),
        plan=plan,
        scorer=model_graded_fact(),
    )




# ruff: noqa: F401 F403 F405

from importlib.metadata import version as importlib_version

from inspect_ai._eval.eval import eval, eval_async, eval_retry, eval_retry_async
from inspect_ai._eval.list import list_tasks
from inspect_ai._eval.registry import task
from inspect_ai._eval.score import score, score_async
from inspect_ai._eval.task import Task, TaskInfo, Tasks
from inspect_ai._util.constants import PKG_NAME

__version__ = importlib_version(PKG_NAME)


__all__ = [
    "__version__",
    "eval",
    "eval_async",
    "eval_retry",
    "eval_retry_async",
    "score",
    "score_async",
    "Task",
    "TaskInfo",
    "Tasks",
    "task",
    "list_tasks",
]



import logging
import re
from random import Random

from inspect_ai._util.pattern import (
    ANSWER_PATTERN_LETTER,
)
from inspect_ai.util import resource

from ._solver import Generate, Solver, TaskState, solver

logger = logging.getLogger(__name__)

# this template is based on the multiple choice template in openai simple evals:
# https://github.com/openai/simple-evals/blob/main/mmlu_eval.py


MULTIPLE_CHOICE_TEMPLATE = r"""
Answer the following multiple choice question. The entire content of your response should be of the following format: 'ANSWER: $LETTER' (without quotes) where LETTER is one of {letters}.

{question}

{choices}
""".strip()


MULTIPLE_CHOICE_TEMPLATE_COT = r"""
Answer the following multiple choice question. The last line of your response should be of the following format: 'ANSWER: $LETTER' (without quotes) where LETTER is one of {letters}. Think step by step before answering.

{question}

{choices}
"""


# max tokens for differnet variations
MULTIPLE_CHOICE_MAX_TOKENS = 32
MULTIPLE_CHOICE_MAX_TOKENS_COT = 1024


@solver
def multiple_choice(
    *,
    cot: bool = False,
    template: str | None = None,
    max_tokens: int | None = None,
    shuffle: bool | Random = False,
    answer_pattern: str | None = None,
) -> Solver:
    """Multiple choice question solver.

    Formats a multiple choice question prompt, then calls `generate()`
    (so you don't need to call `generate()` separately after this solver runs).

    The `template` and `max_tokens` parameters have defaults that vary based
    on whether `cot` is `True`. When NOT using chain of thought,
    `max_tokens` is set to 32 (otherwise it is set to 1024). If you provide your
    own template, you will also need to determine an appropriate value for
    `max_tokens` (as well as `answer_pattern` if `shuffle` is `True`).

    If shuffling is requested, then the choices will be presented in random order,
    and the model output mapped back to the correct choices from the dataset.
    When shuffling is enabled, you must also provide an `answer_pattern` that
    allows this substitution to find the answer in the model output.

    Args:
        cot (bool): `True` to use chain of thought prompting (defaults to `False`).
          Note that using chain of thought will be slower and use more tokens,
          so you should assess carefully whether your eval benefits from it or not.
        template (str | None): Alternate prompt template for questions/answers.
          Templates have 3 variables: `letters`, `question`, and `choices
          (where letters is e.g. 'ABCD').
        max_tokens (int | None): Maximum number of tokens to output.
        shuffle (Random | None): Present answers in a shuffled order (defaults to
          `False`, pass `True` or an instance of `Random` to shuffle)
        answer_pattern (str | None): Regex used to find the answer letter. This is
          only used when `shuffle` is enabled. The regex should have 3 capture groups
         (before the answer, the answer, and after the answer). If the answer is
         expected at the beginning or end then you can use explicit capture groups
         for beginning or end of string, for example (^.*) or (.*$).
    """
    # resolve parameters
    template = (
        template
        if template
        else MULTIPLE_CHOICE_TEMPLATE_COT if cot else MULTIPLE_CHOICE_TEMPLATE
    )
    max_tokens = (
        max_tokens
        if max_tokens
        else MULTIPLE_CHOICE_MAX_TOKENS_COT if cot else MULTIPLE_CHOICE_MAX_TOKENS
    )
    answer_pattern = answer_pattern if answer_pattern else ANSWER_PATTERN_LETTER

    # resolve template contents
    template = resource(template)

    # resolve shuffle
    if shuffle is True:
        shuffle = Random()

    async def solve(state: TaskState, generate: Generate) -> TaskState:
        # confirm we have choices
        if not state.choices:
            raise ValueError("The multiple choice solver requires samples with choices")

        # resolve letters
        letters = "".join(chr(65 + i) for i in range(len(state.choices)))

        # build choices str, key, and prompt

        # unshuffled version (this is what we'll write into history)
        choices_str, _ = make_choices(choices=state.choices)
        user_prompt_text = template.format(
            letters=letters,
            question=state.user_prompt.text,
            choices=choices_str,
        )

        # shuffled version (this is what we'll present to the model)
        choices_str_shuffled, choices_key = make_choices(
            choices=state.choices, shuffle=shuffle if shuffle else None
        )
        state.user_prompt.text = template.format(
            letters=letters,
            question=state.user_prompt.text,
            choices=choices_str_shuffled,
        )

        # generate
        state = await generate(state, max_tokens=max_tokens)

        # unshuffle if necessary
        if shuffle:
            state.output.completion = re.sub(
                answer_pattern,
                lambda m: f"{m.group(1)}{choices_key.get(m.group(2), '')}{m.group(3)}",
                state.output.completion,
            )

        # update last message and restore user prompt
        state.messages[-1].content = state.output.completion
        state.user_prompt.text = user_prompt_text

        # return state
        return state

    return solve


def make_choices(
    choices: list[str],
    shuffle: Random | None = None,
) -> tuple[str, dict[str, str]]:
    # helper to go from index to char
    def answer_char(index: int) -> str:
        return chr(ord("A") + index)

    # shuffle if requested
    indexes = list(range(len(choices)))
    if shuffle:
        shuffle.shuffle(indexes)

    # build choices
    choices_str = "\n".join(
        [f"{answer_char(i)}) {choices[j]}" for i, j in enumerate(indexes)]
    )

    # build key for going from randomized letter to actual label
    choices_key = dict(
        zip(
            [answer_char(i) for i in range(0, len(indexes))],
            [answer_char(i) for i in indexes],
        )
    )

    # return
    return choices_str, choices_key



"""
Measuring Mathematical Problem Solving With the MATH Dataset

Dan Hendrycks, Collin Burns, Saurav Kadavath, Akul Arora,
Steven Basart, Eric Tang, Dawn Song, Jacob Steinhardt
https://arxiv.org/abs/2103.03874

Based on: https://github.com/openai/simple-evals/blob/main/math_eval.py
"""

import re

from inspect_ai import Task, task
from inspect_ai.dataset import FieldSpec, csv_dataset
from inspect_ai.model import GenerateConfig, get_model
from inspect_ai.scorer import (
    CORRECT,
    INCORRECT,
    AnswerPattern,
    Score,
    Target,
    accuracy,
    bootstrap_std,
    scorer,
)
from inspect_ai.solver import TaskState, generate, prompt_template

# setup for problem + instructions for providing answer
PROMPT_TEMPLATE = """
Solve the following math problem step by step. The last line of your response should be of the form "ANSWER: $ANSWER" (without quotes) where $ANSWER is the answer to the problem.

{prompt}

Remember to put your answer on its own line at the end in the form "ANSWER: $ANSWER" (without quotes) where $ANSWER is the answer to the problem, and you do not need to use a \\boxed command.
""".strip()


@task
def math(shuffle=True):
    return Task(
        dataset=csv_dataset(
            csv_file="datasets/math_test.csv",
            sample_fields=FieldSpec(input="Question", target="Answer"),
            shuffle=shuffle,
        ),
        plan=[
            prompt_template(PROMPT_TEMPLATE),
            generate(),
        ],
        scorer=expression_equivalance(),
        config=GenerateConfig(temperature=0.5),
    )


@scorer(metrics=[accuracy(), bootstrap_std()])
def expression_equivalance():
    async def score(state: TaskState, target: Target):
        # extract answer
        match = re.search(AnswerPattern.LINE, state.output.completion)
        if match:
            # ask the model to judge equivalance
            answer = match.group(1)
            prompt = EQUIVALANCE_TEMPLATE % (
                {"expression1": target.text, "expression2": answer}
            )
            result = await get_model().generate(prompt)

            # return the score
            correct = result.completion.lower() == "yes"
            return Score(
                value=CORRECT if correct else INCORRECT,
                answer=answer,
                explanation=state.output.completion,
            )
        else:
            return Score(
                value=INCORRECT,
                explanation="Answer not found in model output: "
                + f"{state.output.completion}",
            )

    return score


EQUIVALANCE_TEMPLATE = r"""
Look at the following two expressions (answers to a math problem) and judge whether they are equivalent. Only perform trivial simplifications

Examples:

  Expression 1: $2x+3$
  Expression 2: $3+2x$

Yes

  Expression 1: 3/2
  Expression 2: 1.5

Yes

  Expression 1: $x^2+2x+1$
  Expression 2: $y^2+2y+1$

No

  Expression 1: $x^2+2x+1$
  Expression 2: $(x+1)^2$

Yes

  Expression 1: 3245/5
  Expression 2: 649

No
(these are actually equal, don't mark them equivalent if you need to
do nontrivial simplifications)

  Expression 1: 2/(-3)
  Expression 2: -2/3

Yes
(trivial simplifications are allowed)

  Expression 1: 72 degrees
  Expression 2: 72

Yes
(give benefit of the doubt to units)

  Expression 1: 64
  Expression 2: 64 square feet

Yes
(give benefit of the doubt to units)

---

YOUR TASK


Respond with only "Yes" or "No" (without quotes). Do not include a rationale.

  Expression 1: %(expression1)s
  Expression 2: %(expression2)s
""".strip()



import base64
import mimetypes

import httpx

from .file import file
from .url import (
    data_uri_mime_type,
    data_uri_to_base64,
    is_data_uri,
    is_http_url,
)


async def image_as_data(image: str) -> tuple[bytes, str]:
    if is_data_uri(image):
        # resolve mime type and base64 content
        mime_type = data_uri_mime_type(image) or "image/png"
        image_base64 = data_uri_to_base64(image)
        image_bytes = base64.b64decode(image_base64)
    else:
        # guess mime type
        type, _ = mimetypes.guess_type(image)
        if type:
            mime_type = type
        else:
            mime_type = "image/png"

        # handle url or file
        if is_http_url(image):
            client = httpx.AsyncClient()
            image_bytes = (await client.get(image)).content
        else:
            with file(image, "rb") as f:
                image_bytes = f.read()

    # return bytes and type
    return image_bytes, mime_type


async def image_as_data_uri(image: str) -> str:
    bytes, mime_type = await image_as_data(image)
    base64_image = base64.b64encode(bytes).decode("utf-8")
    image = f"data:{mime_type};base64,{base64_image}"
    return image



import os
from pathlib import Path

import pytest

from inspect_ai.util import subprocess


@pytest.mark.asyncio
async def test_subprocess_execute():
    result = await subprocess(["python3", "-c", "print('foo')"])
    assert result.stdout.strip() == "foo"


@pytest.mark.asyncio
async def test_suprocess_fail():
    result = await subprocess(["python4"])
    assert result.success is False


@pytest.mark.asyncio
async def test_suprocess_stdin():
    input = "tell me a story"
    result = await subprocess(
        ["python3", "-c", "import sys; print(sys.stdin.read())"], input=input
    )
    assert result.stdout.strip() == input


@pytest.mark.asyncio
async def test_suprocess_binary():
    input = "tell me a story".encode()
    result = await subprocess(
        ["python3", "-c", "import sys; print(sys.stdin.read())"],
        text=False,
        input=input,
    )
    assert result.stdout.decode().strip() == input.decode()


@pytest.mark.asyncio
async def test_subprocess_cwd():
    parent_dir = Path(os.getcwd()).parent.as_posix()
    result = await subprocess(
        ["python3", "-c", "import os; print(os.getcwd())"], cwd=parent_dir
    )
    assert result.stdout.strip() == parent_dir


@pytest.mark.asyncio
async def test_subprocess_env():
    ENV_VAR = "TEST_SUBPROCESS_ENV"
    ENV_VALUE = "test value"
    result = await subprocess(
        ["python3", "-c", f"import os; print(os.getenv('{ENV_VAR}'))"],
        env={ENV_VAR: ENV_VALUE},
    )
    assert result.stdout.strip() == ENV_VALUE


@pytest.mark.asyncio
async def test_subprocess_timeout():
    result = await subprocess(["sleep", "2"], timeout=1)
    assert result.returncode == 1



import re

from inspect_ai.model import ChatMessageUser, Model, get_model
from inspect_ai.solver import TaskState
from inspect_ai.util import resource

from ._metric import INCORRECT, Score
from ._metrics import accuracy, bootstrap_std
from ._scorer import Scorer, Target, scorer


@scorer(metrics=[accuracy(), bootstrap_std()])
def model_graded_fact(
    template: str | None = None,
    instructions: str | None = None,
    grade_pattern: str | None = None,
    partial_credit: bool = False,
    model: str | Model | None = None,
) -> Scorer:
    """Score a question/answer task with a fact response using a model.

    Args:
      template (str): Template for grading prompt. This template uses
        four variables: `question`, `criterion`, `answer`, and
        `instructions` (which is fed from the `instructions` parameter)
      instructions (str): Grading instructions. This should
        include a prompt for the model to answer (e.g. with
        with chain of thought reasoning) in a way that matches
        the specified `grade_pattern`, for example, the default
        `grade_pattern` looks for one of GRADE: C, GRADE: P, or
        GRADE: I).
      grade_pattern (str): Regex to extract the grade from the
        model response. Defaults to looking for e.g. GRADE: C
        The regex should have a single capture group that
        extracts exactly the letter C, P, or I.
      partial_credit (bool): Whether to allow for "partial" credit for
         answers (by default assigned a score of 0.5). Defaults
         to `False`. Note that this parameter is only used
         with the default `instructions` (as custom instructions
         provide their own prompts for grades).
      model (str | Model | none): Model to use for grading
         (by default the model being evaluated is used).
    """
    return model_graded_qa(
        template=template if template else DEFAULT_MODEL_GRADED_FACT_TEMPLATE,
        instructions=instructions,
        grade_pattern=grade_pattern,
        partial_credit=partial_credit,
        model=model,
    )


@scorer(metrics=[accuracy(), bootstrap_std()])
def model_graded_qa(
    template: str | None = None,
    instructions: str | None = None,
    grade_pattern: str | None = None,
    partial_credit: bool = False,
    model: str | Model | None = None,
) -> Scorer:
    """Score a question/answer task using a model.

    Args:
      template (str): Template for grading prompt. This template uses
        four variables: `question`, `criterion`, `answer`, and
        `instructions` (which is fed from the `instructions` parameter)
      instructions (str): Grading instructions. This should
        include a prompt for the model to answer (e.g. with
        with chain of thought reasoning) in a way that matches
        the specified `grade_pattern`, for example, the default
        `grade_pattern` looks for one of GRADE: C, GRADE: P, or
        GRADE: I.
      grade_pattern (str): Regex to extract the grade from the
        model response. Defaults to looking for e.g. GRADE: C
        The regex should have a single capture group that
        extracts exactly the letter C, P, I.
      partial_credit (bool): Whether to allow for "partial" credit for
        answers (by default assigned a score of 0.5). Defaults
        to `False`. Note that this parameter is only used
        with the default `instructions` (as custom instructions
        provide their own prompts for grades).
      model (str | Model | None): Model to use for grading
        (by default the model being evaluated is used).
    """
    # resolve model
    grader_model = get_model(model)

    # resolve grading template, instructions, and grade_pattern
    template = template if template else DEFAULT_MODEL_GRADED_QA_TEMPLATE
    grading_template = resource(template)
    instructions = (
        instructions if instructions else default_instructions(partial_credit)
    )
    grade_pattern = grade_pattern if grade_pattern else DEFAULT_GRADE_PATTERN

    async def score(state: TaskState, target: Target) -> Score:
        # format the scoring template
        score_prompt = grading_template.format(
            question=state.input_text,
            answer=state.output.completion,
            criterion=target.text,
            instructions=instructions,
        )

        # query the model for the score
        result = await grader_model.generate(score_prompt)

        # extract the grade
        match = re.search(grade_pattern, result.completion)
        if match:
            return Score(
                value=match.group(1),
                answer=state.output.completion,
                explanation=result.completion,
                metadata=dict(
                    grading=[
                        ChatMessageUser(content=score_prompt),
                        result.choices[0].message,
                    ]
                ),
            )
        else:
            return Score(
                value=INCORRECT,
                explanation="Grade not found in model output: "
                + f"{result.completion}",
            )

    return score


# these templates are based on the openai closedqa templates here:
# https://github.com/openai/evals/blob/main/evals/registry/modelgraded/closedqa.yaml
DEFAULT_MODEL_GRADED_QA_TEMPLATE = """
You are assessing a submitted answer on a given task based on a criterion. Here is the data:

[BEGIN DATA]
***
[Task]: {question}
***
[Submission]: {answer}
***
[Criterion]: {criterion}
***
[END DATA]

Does the submission meet the criterion?

{instructions}
"""

DEFAULT_MODEL_GRADED_FACT_TEMPLATE = """
You are comparing a submitted answer to an expert answer on a given question. Here is the data:

[BEGIN DATA]
************
[Question]: {question}
************
[Expert]: {criterion}
************
[Submission]: {answer}
************
[END DATA]

Compare the factual content of the submitted answer with the expert answer. Ignore any differences in style, grammar, or punctuation.

Does the submission contain the content in the expert answer?

{instructions}
"""


def default_instructions(partial_credit: bool) -> str:
    partial_letter = "P" if partial_credit else ""
    partial_prompt = '"P" for partially correct answers,' if partial_credit else ""
    return f"""
After assessing the submitted answer, reply with 'GRADE: $LETTER' (without quotes) where LETTER is one of C{partial_letter}I.  Please choose ONE option for the grade: either "C" for correct answers, {partial_prompt}or "I" for incorrect answers.

For example, after reviewing a correct answer you might write 'GRADE: C' or after reviewing an incorrect answer you might write 'GRADE: I'.

First, write out in a step by step manner your reasoning about the criterion to be sure that your conclusion is correct. Avoid simply stating the correct answers at the outset. Then, end with your answer formatted as 'GRADE: $LETTER' (without quotes) where LETTER is one of C{partial_letter}I.
"""


DEFAULT_GRADE_PATTERN = r"(?i)GRADE\s*:\s*([CPI])(.*)$"
"""Regex to extract the grade from the COT above."""



import abc
import asyncio
import os
import sys
import traceback
from logging import LogRecord
from types import TracebackType
from typing import Any, Literal, Type, cast

import click
import tenacity
from pydantic import BaseModel, ConfigDict, Field
from rich.console import Console, RenderableType
from rich.traceback import Traceback

from inspect_ai._util.constants import PKG_NAME
from inspect_ai._util.error import exception_message
from inspect_ai.model import (
    ChatMessage,
    GenerateConfig,
    ModelOutput,
    ModelUsage,
)
from inspect_ai.scorer import Score


class EvalConfig(BaseModel):
    limit: int | tuple[int, int] | None = Field(default=None)
    """Sample limit (number of samples or range of samples)."""

    epochs: int | None = Field(default=None)
    """Number of epochs to run samples over."""

    max_messages: int | None = Field(default=None)
    """Maximum messages to allow in a chat conversation."""

    max_subprocesses: int | None = Field(default=None)
    """Maximum number of subprocesses to run concurrently."""

    log_samples: bool | None = Field(default=None)
    """Log detailed information on each sample."""

    log_images: bool | None = Field(default=None)
    """Log base64 encoded versions of images."""


class EvalSample(BaseModel):
    id: int | str
    """Unique id for sample."""

    epoch: int
    """Epoch number for sample."""

    input: str | list[ChatMessage]
    """Sample input."""

    choices: list[str] | None = Field(default=None)
    """Sample choices."""

    target: str | list[str]
    """Sample target value(s)"""

    messages: list[ChatMessage]
    """Chat conversation history for sample."""

    output: ModelOutput
    """Model output from sample."""

    score: Score | None = Field(default=None)
    """Score for sample."""

    metadata: dict[str, Any]
    """Additional sample metadata."""


class EvalPlanStep(BaseModel):
    solver: str
    """Name of solver."""

    params: dict[str, Any] = Field(default={})
    """Parameters used to instantiate solver."""


class EvalScorer(BaseModel):
    name: str
    """Scorer name."""

    params: dict[str, Any] = Field(default={})
    """Parameters specified when creating scorer."""

    metadata: dict[str, Any] | None = Field(default=None)
    """Additional scorer metadata."""


class EvalPlan(BaseModel):
    name: str = Field(default="plan")
    """Plan name."""

    steps: list[EvalPlanStep] = Field(default=[])
    """Steps in plan."""

    finish: EvalPlanStep | None = Field(default=None)
    """Step to always run at the end."""

    config: GenerateConfig = Field(default=GenerateConfig())
    """Generation config."""


class EvalMetric(BaseModel):
    name: str
    """Metric name."""

    value: int | float
    """Metric value."""

    options: dict[str, Any] = Field(default={})
    """Options specified when creating metric."""

    metadata: dict[str, Any] | None = Field(default=None)
    """Additional metadata associated with metric."""


class EvalResults(BaseModel):
    scorer: EvalScorer | None = Field(default=None)
    """Scorer used to compute results"""

    metrics: dict[str, EvalMetric] = Field(default={})
    """Metrics computed."""

    metadata: dict[str, Any] | None = Field(default=None)
    """Additional results metadata."""


class EvalDataset(BaseModel):
    name: str | None = Field(default=None)
    """Dataset name."""

    location: str | None = Field(default=None)
    """Dataset location (file path or remote URL)"""


class EvalRevision(BaseModel):
    type: Literal["git"]
    """Type of revision (currently only "git")"""

    origin: str
    """Revision origin server"""

    commit: str
    """Revision commit."""


class EvalSpec(BaseModel):
    task: str
    """Task name."""

    task_version: int = Field(default=0)
    """Task version."""

    task_file: str | None = Field(default=None)
    """Task source file."""

    task_id: str = Field(default="")
    """Unique task id."""

    run_id: str = Field(default="")
    """Unqiue run id"""

    created: str
    """Time created."""

    dataset: EvalDataset
    """Dataset used for eval."""

    model: str
    """Model used for eval."""

    model_base_url: str | None = Field(default=None)
    """Optional override of model base url"""

    task_attribs: dict[str, Any] = Field(default={})
    """Attributes of the @task decorator."""

    task_args: dict[str, Any] = Field(default={})
    """Arguments used for involing the task."""

    model_args: dict[str, Any] = Field(default={})
    """Model specific arguments."""

    config: EvalConfig
    """Configuration values for eval."""

    revision: EvalRevision | None = Field(default=None)
    """Source revision of eval."""

    packages: dict[str, str] = Field(default={})
    """Package versions for eval."""

    metadata: dict[str, Any] | None = Field(default=None)
    """Additional eval metadata."""

    # allow field model_args
    model_config = ConfigDict(protected_namespaces=())


class EvalError(BaseModel):
    message: str
    """Error message."""

    traceback: str
    """Error traceback."""

    traceback_ansi: str
    """Error traceback with ANSI color codes."""


def eval_error(
    exception: BaseException,
    exc_type: Type[Any],
    exc_value: BaseException,
    exc_traceback: TracebackType | None,
) -> EvalError:
    # get text traceback
    traceback_text = "\n".join(
        traceback.format_exception(exc_type, exc_value, exc_traceback)
    )

    with open(os.devnull, "w") as f:
        console = Console(record=True, file=f)
        console.print(rich_traceback(exc_type, exc_value, exc_traceback))
        traceback_ansi = console.export_text(styles=True)

    # return error
    return EvalError(
        message=exception_message(exception),
        traceback=traceback_text,
        traceback_ansi=traceback_ansi,
    )


def rich_traceback(
    exc_type: Type[Any], exc_value: BaseException, exc_traceback: TracebackType | None
) -> RenderableType:
    rich_tb = Traceback.from_exception(
        exc_type=exc_type,
        exc_value=exc_value,
        traceback=exc_traceback,
        suppress=[click, asyncio, tenacity, sys.modules[PKG_NAME]],
        show_locals=True,
        max_frames=10,
    )
    return rich_tb


class EvalStats(BaseModel):
    started_at: str = Field(default="")
    """Evaluation start time."""

    completed_at: str = Field(default="")
    """Evaluation completion time."""

    model_usage: dict[str, ModelUsage] = Field(default={})
    """Model token usage for evaluation."""

    # allow field model_usage
    model_config = ConfigDict(protected_namespaces=())


LoggingLevel = Literal["debug", "http", "info", "warning", "error", "critical"]
"""Logging level."""


class LoggingMessage(BaseModel):
    level: LoggingLevel
    """Logging level."""

    message: str
    """Log message."""

    created: float
    """Message created time."""

    @staticmethod
    def from_log_record(record: LogRecord) -> "LoggingMessage":
        """Create a LoggingMesssage from a LogRecord.

        Args:
          record (LogRecord): LogRecord to convert.

        Returns:
          LoggingMessage for LogRecord

        """
        return LoggingMessage(
            level=cast(LoggingLevel, record.levelname.lower()),
            message=record.getMessage(),
            created=record.created * 1000,
        )


class EvalLog(BaseModel):
    version: int = Field(default=1)
    """Eval log file format version."""

    status: Literal["started", "success", "error"] = Field(default="started")
    """Status of evaluation (did it succeed or fail)."""

    eval: EvalSpec
    """Eval identity and configuration."""

    plan: EvalPlan = Field(default=EvalPlan())
    """Eval plan (sovers and config)"""

    results: EvalResults | None = None
    """Eval results (scores and metrics)."""

    stats: EvalStats = Field(default=EvalStats())
    """Eval stats (runtime, model usage)"""

    error: EvalError | None = Field(default=None)
    """Error that halted eval (if status=="error")"""

    samples: list[EvalSample] | None = Field(default=None)
    """Samples processed by eval."""

    logging: list[LoggingMessage] = Field(default=[])
    """Logging message captured during eval."""


LogEvent = Literal["plan", "sample", "score", "results", "scorer", "logging"]


class Recorder(abc.ABC):
    @abc.abstractmethod
    def log_start(self, eval: EvalSpec) -> str:
        pass

    @abc.abstractmethod
    def log_event(
        self,
        spec: EvalSpec,
        type: LogEvent,
        data: EvalSample | EvalPlan | EvalResults | LoggingMessage,
    ) -> None:
        pass

    @abc.abstractmethod
    def log_success(self, eval: EvalSpec, stats: EvalStats) -> EvalLog:
        pass

    @abc.abstractmethod
    def log_failure(
        self, eval: EvalSpec, stats: EvalStats, error: EvalError
    ) -> EvalLog:
        pass

    @abc.abstractmethod
    def read_log(self, location: str) -> EvalLog:
        pass

    @abc.abstractmethod
    def write_log(self, location: str, log: EvalLog) -> None:
        pass

    @abc.abstractmethod
    def read_latest_log(self) -> EvalLog:
        pass



import io
import sys
import types
from pathlib import Path
from typing import Callable

from IPython import get_ipython  # type: ignore
from IPython.core.interactiveshell import InteractiveShell
from nbformat import NBFormatError, ValidationError, read
from nbformat.reader import NotJSONError

# from https://jupyter-notebook.readthedocs.io/en/stable/examples/Notebook/Importing%20Notebooks.html


class NotebookLoader(object):
    """Module Loader for Jupyter Notebooks"""

    def __init__(self, exec_filter: Callable[[list[str]], bool] | None = None) -> None:
        self.shell = InteractiveShell.instance()
        self.exec_filter = exec_filter

    def load_module(self, fullname: str) -> types.ModuleType:
        # load the notebook object
        with io.open(fullname, "r", encoding="utf-8") as f:
            nb = read(f, 4)  # type: ignore

        # create the module and add it to sys.modules
        # if name in sys.modules:
        #    return sys.modules[name]
        mod = types.ModuleType(fullname)
        mod.__file__ = fullname
        mod.__loader__ = self
        mod.__dict__["get_ipython"] = get_ipython
        sys.modules[fullname] = mod

        # extra work to ensure that magics that would affect the user_ns
        # actually affect the notebook module's ns
        save_user_ns = self.shell.user_ns
        self.shell.user_ns = mod.__dict__

        try:
            # get source code for all the calls
            cells_code: list[str] = []
            for cell in nb.cells:
                # transform the input to executable Python for each cell
                if cell.cell_type == "code":
                    code = self.shell.input_transformer_manager.transform_cell(
                        cell.source
                    )
                    cells_code.append(code)

            # check the exec filter to make sure we should execute the
            # notebook cells, if not just return an empty module
            if self.exec_filter and not self.exec_filter(cells_code):
                del sys.modules[fullname]
                return mod

            # run the code in each cell
            for code in cells_code:
                exec(code, mod.__dict__)

            return mod
        finally:
            self.shell.user_ns = save_user_ns


def read_notebook_code(path: Path) -> str:
    try:
        # load the notebook object
        with io.open(path, "r", encoding="utf-8") as f:
            nb = read(f, 4)  # type: ignore
    except NotJSONError:
        return ""
    except ValidationError:
        return ""
    except NBFormatError:
        return ""

    # for dealing w/ magics
    shell = InteractiveShell.instance()

    # get the code
    lines: list[str] = []
    for cell in nb.cells:
        # transform the input to executable Python for each cell
        if cell.cell_type == "code":
            code = shell.input_transformer_manager.transform_cell(cell.source)
            lines.append(code)
    return "\n".join(lines)



import atexit
import json
import logging
import os
import sys
from functools import partial
from http import HTTPStatus
from http.server import HTTPServer
from io import BytesIO
from pathlib import Path
from typing import Any
from urllib.parse import parse_qs, urlparse, urlunparse

import psutil

from inspect_ai._display import display
from inspect_ai._display.logger import init_logger
from inspect_ai._util.appdirs import inspect_runtime_dir
from inspect_ai._util.constants import (
    DEFAULT_SERVER_HOST,
    DEFAULT_VIEW_PORT,
)
from inspect_ai._util.dotenv import init_dotenv
from inspect_ai._util.error import exception_message
from inspect_ai._util.file import FileSystem, file, filesystem
from inspect_ai._util.http import InspectHTTPRequestHandler
from inspect_ai.log._file import eval_log_json, list_eval_logs, read_eval_log

logger = logging.getLogger(__name__)


WWW_DIR = os.path.abspath((Path(__file__).parent / "www").as_posix())


LOGS_PATH = "/api/logs"
LOGS_DIR = f"{LOGS_PATH}/"


def view(
    log_dir: str | None = None,
    recursive: bool = True,
    host: str = DEFAULT_SERVER_HOST,
    port: int = DEFAULT_VIEW_PORT,
    log_level: str | None = None,
    fs_options: dict[str, Any] = {},
) -> None:
    init_dotenv()
    init_logger(log_level)

    # intialize the right filesytem for this log_dir
    log_dir = log_dir if log_dir else os.getenv("INSPECT_LOG_DIR", "./logs")
    fs = filesystem(log_dir, fs_options)

    # list the logs and confirm that there are logs to view (this also ensures
    # that the right e.g. S3 credentials are present before we run the server)
    files = list_eval_logs(log_dir, recursive=recursive, fs_options=fs_options)
    if len(files) == 0:
        print(f"No log files currently available in {log_dir}")
        sys.exit(0)

    # acquire the requested port
    view_acquire_port(port)

    # run server
    view_handler = partial(
        ViewHTTPRequestHandler,
        fs=fs,
        log_dir=log_dir,
        recursive=recursive,
        fs_options=fs_options,
    )
    httpd = HTTPServer((host, port), view_handler)
    display().print(f"Inspect view running at http://localhost:{port}/")
    httpd.serve_forever()


class ViewHTTPRequestHandler(InspectHTTPRequestHandler):
    def __init__(
        self,
        *args: Any,
        fs: FileSystem,
        log_dir: str,
        recursive: bool,
        fs_options: dict[str, Any],
        **kwargs: Any,
    ) -> None:
        self.fs = fs
        self.log_dir = log_dir
        self.recursive = recursive
        self.fs_options = fs_options
        super().__init__(*args, directory=WWW_DIR, **kwargs)

    def do_GET(self) -> None:
        if self.path == LOGS_PATH:
            self.handle_logs()
        elif self.path.startswith(LOGS_DIR):
            self.handle_log()
        else:
            super().do_GET()

    def handle_logs(self) -> None:
        """Serve log files listing from /logs/."""
        files = list_eval_logs(
            self.log_dir, recursive=self.recursive, fs_options=self.fs_options
        )
        json_files = json.dumps(
            dict(
                log_dir=self.log_dir_aliased(),
                files=[
                    dict(
                        name=file.name,
                        size=file.size,
                        mtime=file.mtime,
                        task=file.task,
                        task_id=file.task_id,
                    )
                    for file in files
                ],
                indent=2,
            )
        )
        self.send_json(json_files)

    def handle_log(self) -> None:
        """Serve log files from /api/logs/* url."""
        path = self.path.replace(LOGS_DIR, "", 1)  # strip /api/logs/
        path = path.replace("..", "")  # no escape

        # check for query params
        parsed = urlparse(path)

        # read query parameters from the URL
        query_params = parse_qs(parsed.query)
        header_only = query_params.get("header-only", None) is not None

        # reconstruct the path
        path = urlunparse(
            (
                parsed.scheme,
                parsed.netloc,
                parsed.path,
                parsed.params,
                "",  # Clear the query component
                parsed.fragment,
            )
        )

        ctype = self.guess_type(path)
        try:
            contents: bytes | None = None
            if header_only:
                try:
                    log = read_eval_log(path, header_only=True)
                    contents = eval_log_json(log).encode()
                except ValueError as ex:
                    logger.info(
                        f"Unable to read headers from log file {path}: {exception_message(ex)}. "
                        + "The file may include a NaN or Inf value. Falling back to reading entire file."
                    )
                    pass

            if contents is None:  # normal read
                with file(path, "rb") as f:
                    # read file and determine its length
                    contents = f.read()

            # respond with the log
            length = len(contents)
            self.send_response(HTTPStatus.OK)
            self.send_header("Content-type", ctype)
            self.send_header("Content-Length", str(length))
            self.end_headers()
            self.copyfile(BytesIO(contents), self.wfile)  # type: ignore
        except Exception as error:
            logger.exception(error)
            self.send_error(HTTPStatus.NOT_FOUND, "File not found")

    def events_response(self, params: dict[str, str]) -> list[str]:
        last_eval_time = params.get("last_eval_time", None)
        actions = (
            ["refresh-evals"]
            if last_eval_time and view_last_eval_time() > int(last_eval_time)
            else []
        )
        return super().events_response(params) + actions

    def log_dir_aliased(self) -> str:
        home_dir = os.path.expanduser("~")
        if self.log_dir.startswith(home_dir):
            return self.log_dir.replace(home_dir, "~", 1)
        else:
            return self.log_dir


# lightweight tracking of when the last eval task completed
# this enables the view client to poll for changes frequently
# (e.g. every 1 second) with very minimal overhead.


def view_notify_eval(location: str) -> None:
    file = view_last_eval_file()
    with open(file, "w", encoding="utf-8") as f:
        if not urlparse(location).scheme:
            location = Path(location).absolute().as_posix()
        f.write(location)


def view_last_eval_time() -> int:
    file = view_last_eval_file()
    if file.exists():
        return int(file.stat().st_mtime * 1000)
    else:
        return 0


def view_runtime_dir() -> Path:
    return inspect_runtime_dir("view")


def view_last_eval_file() -> Path:
    return view_runtime_dir() / "last-eval"


def view_port_pid_file(port: int) -> Path:
    ports_dir = view_runtime_dir() / "ports"
    ports_dir.mkdir(parents=True, exist_ok=True)
    return ports_dir / str(port)


def view_acquire_port(port: int) -> None:
    # pid file name
    pid_file = view_port_pid_file(port)

    # does it already exist? if so terminate that process
    if pid_file.exists():
        WAIT_SECONDS = 5
        with open(pid_file, "r", encoding="utf-8") as f:
            pid = int(f.read().strip())
        try:
            p = psutil.Process(pid)
            p.terminate()
            display().print(
                f"Terminating existing inspect view command using port {port}"
            )
            p.wait(WAIT_SECONDS)

        except psutil.NoSuchProcess:
            # expected error for crufty pid files
            pass
        except psutil.TimeoutExpired:
            logger.warning(
                f"Timed out waiting for process to exit for {WAIT_SECONDS} seconds."
            )
        except psutil.AccessDenied:
            logger.warning(
                "Attempted to kill existing view command on "
                + f"port {port} but access was denied."
            )
        except Exception as ex:
            logger.warning(
                "Attempted to kill existing view command on "
                + f"port {port} but error occurred: {exception_message(ex)}"
            )

    # write our pid to the file
    with open(pid_file, "w", encoding="utf-8") as f:
        f.write(str(os.getpid()))

    # arrange to release on exit
    def release_lock_file() -> None:
        try:
            pid_file.unlink(True)
        except Exception:
            pass

    atexit.register(release_lock_file)



def parse_samples_limit(limit: str | None) -> int | tuple[int, int] | None:
    if limit is not None:
        if "," not in limit:
            return int(limit)
        else:
            limit_split = [int(r) for r in limit.split(",")]
            return (limit_split[0] - 1, limit_split[1])
    else:
        return None



import pytest
from utils import skip_if_no_anthropic

from inspect_ai.model import GenerateConfig, get_model


@pytest.mark.asyncio
@skip_if_no_anthropic
async def test_anthropic_api() -> None:
    model = get_model(
        "claude-2.1",
        config=GenerateConfig(
            frequency_penalty=0.0,
            stop_seqs=None,
            max_tokens=50,
            presence_penalty=0.0,
            seed=None,
            temperature=0.0,
            top_p=1.0,
        ),
    )

    message = "This is a test string. What are you?"
    response = await model.generate(input=message)
    assert len(response.completion) >= 1



import asyncio
import re
from copy import deepcopy
from typing import Callable, cast

from inspect_ai._display import display
from inspect_ai._util.platform import platform_init
from inspect_ai._util.registry import (
    registry_create,
    registry_info,
    registry_log_name,
    registry_params,
    registry_unqualified_name,
)
from inspect_ai.log import EvalLog, EvalMetric, EvalResults, EvalScorer
from inspect_ai.model import ModelName
from inspect_ai.scorer import Metric, Score, Scorer, Target
from inspect_ai.scorer._scorer import SCORER_METRICS, scorer_metrics
from inspect_ai.solver import TaskState


def score(log: EvalLog, scorer: Scorer) -> EvalLog:
    """Score an evaluation log.

    Args:
       log (EvalLog): Evaluation log.
       scorer (Scorer): Scorer to apply to log
       metrics: (list[Metric]): Additional metrics to compute
         (Scorer built-in metrics are always computed).

    Returns:
       Log with scores yielded by scorer.
    """
    # standard platform init for top level entry points
    platform_init()

    return asyncio.run(score_async(log, scorer))


async def score_async(log: EvalLog, scorer: Scorer) -> EvalLog:
    """Score an evaluation log.

    Args:
       log (EvalLog): Evaluation log.
       scorer (Scorer): Scorer to apply to log

    Returns:
       Log with scores yielded by scorer.
    """
    # deepcopy so we don't mutate the passed log
    log = deepcopy(log)

    # confirm we have samples
    if log.samples is None or len(log.samples) == 0:
        raise ValueError("There are no samples to score in the log.")

    # prime the scoring tasks
    states = [
        TaskState(
            model=ModelName(log.eval.model),
            sample_id=sample.id,
            epoch=sample.epoch,
            input=sample.input,
            choices=sample.choices,
            messages=sample.messages,
            output=sample.output,
            completed=True,
            metadata=sample.metadata,
        )
        for sample in log.samples
    ]
    with display().progress(total=len(states)) as p:

        def progress() -> None:
            p.update(1)

        tasks = [
            run_score_task(state, Target(sample.target), scorer, progress)
            for (sample, state) in zip(log.samples, states)
        ]

        # do scoring
        scores = await asyncio.gather(*tasks)

        # write them back (gather ensures that they come back in the same order)
        for index, score in enumerate(scores):
            log.samples[index].score = score

        # collect metrics from EvalLog (they may overlap w/ the scorer metrics,
        # that will be taken care of in eval_results)
        log_metrics = metrics_from_log(log)

        # compute metrics
        log.results = eval_results(scores, scorer, log_metrics)

    return log


async def run_score_task(
    state: TaskState,
    target: Target,
    scorer: Scorer,
    progress: Callable[..., None],
) -> Score:
    result = await scorer(state, target)
    progress()
    return result


def eval_results(
    scores: list[Score], scorer: Scorer | None, metrics: list[Metric] = []
) -> EvalResults:
    # record scorer
    results = EvalResults()
    if scorer:
        # extract non-metrics metadata
        metadata = deepcopy(registry_info(scorer).metadata)
        del metadata[SCORER_METRICS]

        # build results
        results.scorer = EvalScorer(
            name=registry_log_name(scorer),
            params=registry_params(scorer),
            metadata=metadata if len(metadata.keys()) > 0 else None,
        )

        # we want to use simple names for metrics in the metrics dict
        # (i.e. without package prefixes). we do this by getting the
        # unqualified name, then appending a suffix if there are duplicates
        # this keeps the code straightforward and intuitive for users
        # programming against the log (e.g. metrics["accuracy"]) vs.
        # metrics["pkgname/accuracy"])
        for metric in target_metrics(scorer, metrics):
            key = metrics_unique_key(
                registry_unqualified_name(metric), list(results.metrics.keys())
            )
            results.metrics[key] = EvalMetric(
                name=registry_log_name(metric), value=metric(scores)
            )
    return results


def metrics_unique_key(key: str, existing: list[str]) -> str:
    if key not in existing:
        return key
    else:
        key_index = 2
        pattern = re.compile(f"{re.escape(key)}(\\d+)")
        for existing_key in existing:
            match = pattern.match(existing_key)
            index = int(match.group(1)) if match else None
            if index and (index >= key_index):
                key_index = index + 1
        return f"{key}{key_index}"


# build a list of metrics (scorer built-in metrics + de-duplicated additional metrics)
def target_metrics(scorer: Scorer, metrics: list[Metric]) -> list[Metric]:
    target_metrics = scorer_metrics(scorer)
    target_metrics_names = [registry_log_name(metric) for metric in target_metrics]
    target_metrics.extend(
        [
            metric
            for metric in metrics
            if registry_log_name(metric) not in target_metrics_names
        ]
    )
    return target_metrics


def metrics_from_log(log: EvalLog) -> list[Metric]:
    return (
        [metric_from_log(metric) for metric in log.results.metrics.values()]
        if log.results
        else []
    )


def metric_from_log(metric: EvalMetric) -> Metric:
    return cast(Metric, registry_create("metric", metric.name, **metric.options))



from typing import Any

from inspect_ai.model import (
    ChatMessage,
    ChatMessageAssistant,
    ChatMessageSystem,
    ChatMessageTool,
    ChatMessageUser,
)

from ._dataset import (
    DatasetRecord,
    FieldSpec,
    RecordToSample,
    Sample,
)


# determine how we will go from file records to samples. if there is
# no field spec, we assume the column names "input" and "target",
# otherwise use the provided field spec or custom converter function
def record_to_sample_fn(
    sample_fields: FieldSpec | RecordToSample | None,
) -> RecordToSample:
    if sample_fields is None:
        sample_fields = FieldSpec()

    if isinstance(sample_fields, FieldSpec):

        def record_to_sample(record: DatasetRecord) -> Sample:
            # collect metadata if specified
            metadata: dict[str, Any] | None = None
            if sample_fields.metadata:
                metadata = {}
                for name in sample_fields.metadata:
                    metadata[name] = record.get(name)

            # return sample
            return Sample(
                input=read_input(record.get(sample_fields.input)),
                target=read_target(record.get(sample_fields.target)),
                choices=read_choices(record.get(sample_fields.choices)),
                id=record.get(sample_fields.id, None),
                metadata=metadata,
            )

    else:

        def record_to_sample(record: DatasetRecord) -> Sample:
            return sample_fields(record)

    return record_to_sample


def read_input(input: Any | None) -> str | list[ChatMessage]:
    if not input:
        raise ValueError("No input in dataset")
    if not isinstance(input, str):
        return read_messages(input)
    else:
        return input


def read_messages(messages: list[dict[str, Any]]) -> list[ChatMessage]:
    chat_messages: list[ChatMessage] = []
    for message in messages:
        role = message.get("role", None)

        content = message.get("content", None)
        if content is None:
            raise ValueError("content not specified for chat input in dataset")

        match role:
            case "system":
                chat_messages.append(ChatMessageSystem(content=content, source="input"))
            case "user":
                chat_messages.append(ChatMessageUser(content=content, source="input"))
            case "assistant":
                chat_messages.append(
                    ChatMessageAssistant(
                        content=content,
                        source="input",
                        tool_calls=message.get("tool_calls", None),
                    )
                )
            case "tool":
                chat_messages.append(
                    ChatMessageTool(
                        content=content,
                        source="input",
                        tool_call_id=message.get("tool_call_id", None),
                        tool_error=message.get("tool_error", None),
                    )
                )
            case _:
                raise ValueError("role not specified for chat input in dataset")

    return chat_messages


def read_target(obj: Any | None) -> str | list[str]:
    if obj is not None:
        return [str(item) for item in obj] if isinstance(obj, list) else str(obj)
    else:
        return ""


def read_choices(obj: Any | None) -> list[str] | None:
    if obj is not None:
        if isinstance(obj, list):
            return [str(choice) for choice in obj]
        elif isinstance(obj, str):
            choices = obj.split(",")
            if len(choices) == 1:
                choices = obj.split()
            return [choice.strip() for choice in choices]
        else:
            return [str(obj)]
    else:
        return None



from inspect_ai.model import (
    ChatMessageUser,
    Model,
    get_model,
)
from inspect_ai.util import resource

from ._solver import Generate, Solver, TaskState, solver


@solver
def self_critique(
    critique_template: str | None = None,
    completion_template: str | None = None,
    model: str | Model | None = None,
) -> Solver:
    """Solver which uses a model to critique the original answer.

    The `critique_template` is used to generate a critique
    and the `completion_template` is used to play that critique
    back to the model for an improved response. Note that you
    can specify an alternate `model` for critique (you don't
    need to use the model being evaluated).

    Args:
      critique_template (str | None): String or path to file
         containing critique template. The template uses two
         variables: `question` and `completion`.
      completion_template (str | None): String or path to file
          containing completion template. The template uses
          three variables: `question`,  `completion`, and `critique`
      model (str | Model | None): Alternate model to be used
         for critique (by default the model being evaluated
         is used).
    """
    # resolve templates
    critique_template = resource(critique_template or DEFAULT_CRITIQUE_TEMPLATE)
    completion_template = resource(
        completion_template or DEFAULT_CRITIQUE_COMPLETION_TEMPLATE
    )

    # resolve model
    model = get_model(model)

    async def solve(state: TaskState, generate: Generate) -> TaskState:
        # run critique
        critique = await model.generate(
            critique_template.format(
                question=state.input_text,
                completion=state.output.completion,
            )
        )

        # add the critique as a user message
        state.messages.append(
            ChatMessageUser(
                content=completion_template.format(
                    question=state.input_text,
                    completion=state.output.completion,
                    critique=critique.completion,
                ),
            )
        )

        # regenerate
        return await generate(state)

    return solve


DEFAULT_CRITIQUE_TEMPLATE = r"""
Given the following question and answer, please critique the answer. A good answer comprehensively answers the question and NEVER refuses to answer. If the answer is already correct do not provide critique - simply respond 'The original answer is fully correct'.

[BEGIN DATA]
***
[Question]: {question}
***
[Answer]: {completion}
***
[END DATA]

Critique: """


DEFAULT_CRITIQUE_COMPLETION_TEMPLATE = r"""
Given the following question, initial answer and critique please generate an improved answer to the question:

[BEGIN DATA]
***
[Question]: {question}
***
[Answer]: {completion}
***
[Critique]: {critique}
***
[END DATA]

If the original answer is already correct, just repeat the original answer exactly. Provide your answer at the end on its own line in the form "ANSWER: $ANSWER" (without quotes) where $ANSWER is the answer to the question.
"""



import csv
from io import TextIOWrapper
from pathlib import Path
from typing import Any

from inspect_ai._util.file import file

from .._dataset import (
    Dataset,
    DatasetReader,
    FieldSpec,
    MemoryDataset,
    RecordToSample,
)
from .._util import record_to_sample_fn


def csv_dataset(
    csv_file: str,
    sample_fields: FieldSpec | RecordToSample | None = None,
    shuffle: bool = False,
    seed: int | None = None,
    limit: int | None = None,
    dialect: str = "unix",
    encoding: str = "utf-8",
    name: str | None = None,
    fs_options: dict[str, Any] = {},
) -> Dataset:
    r"""Read dataset from CSV file.

    Args:
        csv_file (str): Path to CSV file. Can be a local filesystem path or
            a path to an S3 bucket (e.g. "s3://my-bucket"). Use `fs_options`
            to pass arguments through to the `S3FileSystem` constructor.
        sample_fields (SampleFieldSpec | RecordToSample): Method of mapping underlying
            fields in the data source to Sample objects. Pass `None` if the data is already
            stored in `Sample` form (i.e. has "input" and "target" columns.); Pass a
            `SampleFieldSpec` to specify mapping fields by name; Pass a `RecordToSample` to
            handle mapping with a custom function.
        shuffle (bool): Randomly shuffle the dataset order.
        seed: (int | None): Seed used for random shuffle.
        limit (int | None): Limit the number of records to read.
        dialect (str): CSV dialect ("unix" or "excel", defaults to "unix").
        encoding (str): Text encoding for file (defaults to "utf-8").
        name (str): Optional name for dataset (for logging). If not specified,
            defaults to the stem of the filename
        fs_options (dict[str, Any]): Optional. Addional arguments to pass through
            to the filesystem provider (e.g. `S3FileSystem`). Use `{"anon": True }`
            if you are accessing a public S3 bucket with no credentials.

    Returns:
        Dataset read from CSV file.
    """
    # resolve data_to_sample function
    data_to_sample = record_to_sample_fn(sample_fields)

    # read and convert samples
    with file(csv_file, "r", encoding=encoding, fs_options=fs_options) as f:
        # filter out rows with empty values
        valid_data = [
            data
            for data in csv_dataset_reader(f, dialect)
            if data and any(value.strip() for value in data.values())
        ]
        name = name if name else Path(csv_file).stem
        dataset = MemoryDataset(
            samples=[data_to_sample(data) for data in valid_data],
            name=name,
            location=csv_file,
        )

        # shuffle if requested
        if shuffle:
            dataset.shuffle(seed=seed)

        # limit if requested
        if limit:
            dataset = MemoryDataset(list(dataset[0:limit]))

        return dataset


def csv_dataset_reader(file: TextIOWrapper, dialect: str = "unix") -> DatasetReader:
    return csv.DictReader(file, dialect=dialect)



from inspect_ai._util.error import pip_dependency_error
from inspect_ai._util.version import verify_required_version

from .._model import ModelAPI
from .._registry import modelapi

# Defer importing model api classes until they are actually used
# (this allows the package to load without the optional deps)
# Note that some api providers (e.g. CloudFlare, AzureAI) don't
# strictly require this treament but we do it anyway for uniformity,


@modelapi(name="openai", models=["gpt"])
def openai() -> type[ModelAPI]:
    # validate
    validate_openai_client("OpenAI API")

    # in the clear
    from .openai import OpenAIAPI

    return OpenAIAPI


@modelapi(name="anthropic", models=["claude"])
def anthropic() -> type[ModelAPI]:
    FEATURE = "Anthropic API"
    PACKAGE = "anthropic"
    MIN_VERSION = "0.23.0"

    # verify we have the package
    try:
        import anthropic  # noqa: F401
    except ImportError:
        raise pip_dependency_error(FEATURE, [PACKAGE])

    # verify version
    verify_required_version(FEATURE, PACKAGE, MIN_VERSION)

    # in the clear
    from .anthropic import AnthropicAPI

    return AnthropicAPI


@modelapi(name="google", models=["gemini", "bison", "gdm"])
def google() -> type[ModelAPI]:
    FEATURE = "Google API"
    PACKAGE = "google-generativeai"
    MIN_VERSION = "0.4.0"

    # verify we have the package
    try:
        import google.generativeai  # type: ignore  # noqa: F401
    except ImportError:
        raise pip_dependency_error(FEATURE, [PACKAGE])

    # verify version
    verify_required_version(FEATURE, PACKAGE, MIN_VERSION)

    # in the clear
    from .google import GoogleAPI

    return GoogleAPI


@modelapi(name="hf")
def hf() -> type[ModelAPI]:
    try:
        from .hf import HuggingFaceAPI
    except ImportError:
        raise pip_dependency_error("Hugging Face Models", ["torch", "transformers"])

    return HuggingFaceAPI


@modelapi(name="cf")
def cf() -> type[ModelAPI]:
    from .cloudflare import CloudFlareAPI

    return CloudFlareAPI


@modelapi(name="mistral")
def mistral() -> type[ModelAPI]:
    FEATURE = "Mistral API"
    PACKAGE = "mistralai"
    MIN_VERSION = "0.1.3"

    # verify we have the package
    try:
        import mistralai  # noqa: F401
    except ImportError:
        raise pip_dependency_error(FEATURE, [PACKAGE])

    # verify version
    verify_required_version(FEATURE, PACKAGE, MIN_VERSION)

    # in the clear
    from .mistral import MistralAPI

    return MistralAPI


@modelapi(name="together")
def together() -> type[ModelAPI]:
    # validate
    validate_openai_client("TogetherAI API")

    # in the clear
    from .together import TogetherAIAPI

    return TogetherAIAPI


@modelapi(name="azureai")
def azureai() -> type[ModelAPI]:
    from .azureai import AzureAIAPI

    return AzureAIAPI


@modelapi(name="bedrock")
def bedrock() -> type[ModelAPI]:
    from .bedrock import BedrockAPI

    return BedrockAPI


def validate_openai_client(feature: str) -> None:
    FEATURE = feature
    PACKAGE = "openai"
    MIN_VERSION = "1.11.0"

    # verify we have the package
    try:
        import openai  # noqa: F401
    except ImportError:
        raise pip_dependency_error(FEATURE, [PACKAGE])

    # verify version
    verify_required_version(FEATURE, PACKAGE, MIN_VERSION)



from logging import getLogger
from typing import (
    Any,
    Callable,
    Protocol,
    TypeVar,
    Union,
    cast,
    overload,
    runtime_checkable,
)

from pydantic import BaseModel, Field

from inspect_ai._util.registry import (
    RegistryInfo,
    registry_add,
    registry_create,
    registry_name,
    registry_tag,
)

logger = getLogger(__name__)

CORRECT = "C"
"""Value to assign for correct answers."""

INCORRECT = "I"
"""Value to assing for incorrect answers."""

PARTIAL = "P"
"""Value to assign for partial credit."""

NOANSWER = "N"
"""Value to assign for no answer or refusal to answer."""


Value = Union[
    str | int | float | bool,
    list[str | int | float | bool],
    dict[str, str | int | float | bool],
]
"""Value provided by a score.

Use the methods of `Score` to easily treat
the Value as a simple scalar of various types.
"""


class Score(BaseModel):
    """Score generated by a scorer.

    Args:
       value (Value): Score value.
       answer (str | None): Answer extracted from model output (optional).
       explanation (str | None): Explanation of score (optional).
       metadata (dict[str,Any]): Additional metadata related to the score.
    """

    value: Value
    """Score value."""

    answer: str | None = Field(default=None)
    """Answer extracted from model output (optional)"""

    explanation: str | None = Field(default=None)
    """Explanation of score (optional)."""

    metadata: dict[str, Any] | None = Field(default=None)
    """Additional metadata related to the score"""

    @property
    def text(self) -> str:
        """Read the score as text."""
        return self.as_str()

    def as_str(self) -> str:
        """Read the score as a string."""
        return str(self._as_scalar())

    def as_int(self) -> int:
        """Read the score as an integer."""
        return int(self._as_scalar())

    def as_float(self) -> float:
        """Read the score as a float."""
        return float(self._as_scalar())

    def as_bool(self) -> bool:
        """Read the score as a boolan."""
        return bool(self._as_scalar())

    def _as_scalar(self) -> str | int | float | bool:
        if (
            isinstance(self.value, str)
            or isinstance(self.value, int)
            or isinstance(self.value, float)
            or isinstance(self.value, bool)
        ):
            return self.value
        else:
            raise ValueError("This score is not a scalar")


ValueToFloat = Callable[[Value], float]
"""Function used by metrics to translate from a Score value to a float value."""


def value_to_float(
    correct: Value = CORRECT,
    incorrect: Value = INCORRECT,
    partial: Value = PARTIAL,
    noanswer: Value = NOANSWER,
) -> ValueToFloat:
    """Create a ValueToFloat function.

    Create a ValueToFloat function that maps string values of
    the form "C", "I", "P", and "N" to 1, 0, 0.5, and 0
    (respectively). Note that those are the default literal
    values, but they can be customized. Numeric values are
    cast to float. Arrays and dictionaries give a warning
    and return 0.

    Args:
       correct (Value): Value that represents a correct answer (1)
       incorrect (Value): Value that represents an incorrect answer (0)
       partial (Value): Value to assign partial credit for (0.5)
       noanswer (Value): Value for refusals to answer (0)

    Returns:
        ValueToFloat function.
    """

    def to_float(value: Value) -> float:
        if isinstance(value, (int, float, bool)):
            return float(value)
        elif value == correct:
            return 1.0
        elif value == partial:
            return 0.5
        elif value == incorrect or value == noanswer:
            return 0
        else:
            logger.warning(f"Unable to convert value to float: {value}")
            return 0

    return to_float


@runtime_checkable
class Metric(Protocol):
    r"""Evaluate scores using a metric.

    Args:
        scores (list[dict]): List of scores.

    Returns:
        Metric value
    """

    def __call__(self, scores: list[Score]) -> int | float: ...


MetricType = TypeVar("MetricType", Callable[..., Metric], type[Metric])
r"""Metric type.
Valid metric types include:
 - Functions that return a Metric
 - Classes derivied from Metric
"""


def metric_register(metric: MetricType, name: str = "") -> MetricType:
    r"""Register a function or class as a metric.

    Args:
        metric (MetricType):
            Function that returns a Metric or class
            deriving fromMetric
        name (str): Name of metric (Optional, defaults to object name)

    Returns:
        Metric type with registry attributes.
    """
    metric_name = name if name else getattr(metric, "__name__")
    registry_add(metric, RegistryInfo(type="metric", name=metric_name))
    return metric


def metric_create(name: str, **kwargs: Any) -> Metric:
    r"""Create a Metric based on its registered name.

    Metrics can be functions that return a Metric or classes
    deriving from Metric

    Args:
        name (str): Name of metric (Optional, defaults to object name)
        **kwargs (dict): Optional creation arguments for the metric

    Returns:
        Metric with registry info attribute
    """
    return cast(Metric, registry_create("metric", name, **kwargs))


@overload
def metric(name: str) -> Callable[..., MetricType]: ...


@overload
# type: ignore
def metric(name: Callable[..., Metric]) -> Callable[..., Metric]: ...


@overload
def metric(name: type[Metric]) -> type[Metric]: ...


def metric(name: str | MetricType) -> Callable[..., MetricType] | MetricType:
    r"""Decorator for registering metrics.

    Args:
        name: (str | MetricType):
            Optional name for metric. If the decorator has no name
            argument then the name of the underlying MetricType
            will be used to automatically assign a name.
    """

    # create_metric_wrapper:
    #  (a) Add the MetricType to the registry using the appropriately
    #      package-namespaced name
    #  (b) Ensure that instances of Metric created by MetricType also
    #      carry registry info.
    def create_metric_wrapper(
        metric_type: MetricType, name: str | None = None
    ) -> MetricType:
        metric_name = registry_name(
            metric_type, name if name else getattr(metric_type, "__name__")
        )

        def metric_wrapper(*args: Any, **kwargs: Any) -> Metric:
            metric = metric_type(*args, **kwargs)
            registry_tag(
                metric_type,
                metric,
                RegistryInfo(type="metric", name=metric_name),
                *args,
                **kwargs,
            )
            return metric

        return metric_register(cast(MetricType, metric_wrapper), metric_name)

    # for decorators with an explicit name, one more wrapper for the name
    if isinstance(name, str):

        def wrapper(metric_type: MetricType) -> MetricType:
            return create_metric_wrapper(metric_type, name)

        return wrapper

    # create a metric wrapper for the passsed metric_type
    else:
        metric_type = name
        return create_metric_wrapper(metric_type)



import pytest
from utils import skip_if_no_openai

from inspect_ai.model import (
    ChatMessageUser,
    GenerateConfig,
    get_model,
)


@pytest.mark.asyncio
@skip_if_no_openai
async def test_openai_api() -> None:
    model = get_model(
        "openai/gpt-3.5-turbo",
        config=GenerateConfig(
            frequency_penalty=0.0,
            stop_seqs=None,
            max_tokens=50,
            presence_penalty=0.0,
            logit_bias=dict([(42, 10), (43, -10)]),
            seed=None,
            temperature=0.0,
            top_p=1.0,
        ),
    )

    message = ChatMessageUser(content="This is a test string. What are you?")
    response = await model.generate(input=[message])
    assert len(response.completion) >= 1
